#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

from pyscf import lib
from pyscf.dft.gen_grid import Grids

from ciderpress.dft.model_utils import get_slxc_settings, load_cider_model
from ciderpress.pyscf.gen_cider_grid import CiderGrids
from ciderpress.pyscf.nldf_convolutions import PySCFNLDFInitializer
from ciderpress.pyscf.numint import CiderNumInt, NLDFNLOFNumInt, NLDFNumInt, NLOFNumInt
from ciderpress.pyscf.sdmx import PySCFSDMXInitializer


def make_cider_calc(
    ks,
    mlfunc,
    xmix=1.0,
    xc=None,
    xkernel=None,
    ckernel=None,
    mlfunc_format=None,
    nlc_coeff=None,
    nldf_init=None,
    sdmx_init=None,
    rhocut=None,
):
    """
    Decorate the PySCF DFT object ks with a CIDER functional mlfunc.
    If xc, xkernel, ckernel, and xmix are not specified,
    the equivalent of HF with CIDER in place of EXX is performed.
    The XC energy is

    If CIDER is X only:
        E_xc = xmix * E_x^CIDER + (1-xmix) * xkernel + ckernel + xc
    If CIDER is X + C:
        E_xc = xmix * E_xc^CIDER + (1-xmix) * (xkernel + ckernel) + xc

    NOTE: Only GGA-level XC functionals can be used with GGA-level
        (orbital-independent) CIDER functionals currently.

    Args:
        ks (pyscf.dft.KohnShamDFT): DFT object
        mlfunc (MappedXC, MappedXC2, str): CIDER exchange functional or file name
        xmix (float): Fraction of CIDER exchange used.
        xc (str or None): If specified, this semi-local XC code is evaluated
             and added to the total XC energy.
        xkernel (str or None): Semi-local X code in libxc. Scaled by (1-xmix).
        ckernel (str or None): Semi-local C code in libxc.
        mlfunc_format (str or None): 'joblib' or 'yaml', specifies the format
            of mlfunc if it is a string corresponding to a file name.
            If unspecified, infer from file extension and raise error
            if file type cannot be determined.
        nlc_coeff (tuple or None):
            VV10 coefficients. If None, VV10 term is not evaluated.
        nldf_init (PySCFNLDFInitializer)
        sdmx_init (PySCFSDMXInitializer)
        rhocut (float)

    Returns:
        A decorated Kohn-Sham object for performing a CIDER calculation.
    """
    mlfunc = load_cider_model(mlfunc, mlfunc_format)
    ks.xc = get_slxc_settings(xc, xkernel, ckernel, xmix)
    new_ks = _CiderKS(
        ks,
        mlfunc,
        xmix=xmix,
        nldf_init=nldf_init,
        sdmx_init=sdmx_init,
        rhocut=rhocut,
        nlc_coeff=nlc_coeff,
    )
    return lib.set_class(new_ks, (_CiderKS, ks.__class__))


class _CiderKS:

    grids = None

    def __init__(
        self,
        mf,
        mlxc,
        xmix=1.0,
        nldf_init=None,
        sdmx_init=None,
        rhocut=None,
        nlc_coeff=None,
    ):
        self.__dict__.update(mf.__dict__)
        # self.mlxc = None
        # self.xmix = None
        # self.nldf_init = None
        # self.sdmx_init = None
        # self.rhocut = rhocut
        self.set_mlxc(
            mlxc,
            xmix=xmix,
            nldf_init=nldf_init,
            sdmx_init=sdmx_init,
            rhocut=rhocut,
            nlc_coeff=nlc_coeff,
        )

    def set_mlxc(
        self,
        mlxc,
        xmix=1.0,
        nldf_init=None,
        sdmx_init=None,
        rhocut=None,
        nlc_coeff=None,
    ):
        # self.mlxc = mlxc
        # self.xmix = xmix
        # self.nldf_init = nldf_init
        # self.sdmx_init = sdmx_init
        if nldf_init is None and mlxc.settings.has_nldf:
            nldf_init = PySCFNLDFInitializer(mlxc.settings.nldf_settings)
        if sdmx_init is None and mlxc.settings.has_sdmx:
            sdmx_init = PySCFSDMXInitializer(mlxc.settings.sdmx_settings, lowmem=False)
        old_grids = self.grids
        changed = False
        if mlxc.settings.has_nldf:
            if not isinstance(old_grids, CiderGrids):
                changed = True
                self.grids = CiderGrids(self.mol)
        else:
            if old_grids is None or isinstance(old_grids, CiderGrids):
                changed = True
                self.grids = Grids(self.mol)
        if changed:
            for key in (
                "atom_grid",
                "atomic_radii",
                "radii_adjust",
                "radi_method",
                "becke_scheme",
                "prune",
                "level",
            ):
                self.grids.__setattr__(key, old_grids.__getattribute__(key))
        settings = mlxc.settings
        has_nldf = not settings.nldf_settings.is_empty
        has_nlof = not settings.nlof_settings.is_empty
        if has_nldf and has_nlof:
            cls = NLDFNLOFNumInt
        elif has_nldf:
            cls = NLDFNumInt
        elif has_nlof:
            cls = NLOFNumInt
        else:
            cls = CiderNumInt
        self._numint = cls(
            mlxc, nldf_init, sdmx_init, xmix=xmix, rhocut=rhocut, nlc_coeff=nlc_coeff
        )

    def build(self, mol=None, **kwargs):
        self._numint.build(mol=mol)
        return super().build(mol, **kwargs)

    def reset(self, mol=None):
        self._numint.reset(mol=mol)
        return super().reset(mol)

    def method_not_implemented(self, *args, **kwargs):
        raise NotImplementedError

    def density_fit(self, auxbasis=None, with_df=None, only_dfj=False):
        new_self = super().density_fit(
            auxbasis=auxbasis, with_df=with_df, only_dfj=only_dfj
        )
        lib.set_class(new_self, (_CiderDF, new_self.__class__))
        return new_self

    def nuc_grad_method(self):
        from pyscf import dft

        has_df = hasattr(self, "with_df") and self.with_df is not None
        if isinstance(self, dft.rks.RKS):
            from ciderpress.pyscf import rks_grad

            if has_df:
                return rks_grad.DFGradients(self)
            else:
                return rks_grad.Gradients(self)
        elif isinstance(self, dft.uks.UKS):
            from ciderpress.pyscf import uks_grad

            if has_df:
                return uks_grad.DFGradients(self)
            else:
                return uks_grad.Gradients(self)

    Gradients = nuc_grad_method

    Hessian = method_not_implemented
    NMR = method_not_implemented
    NSR = method_not_implemented
    Polarizability = method_not_implemented
    RotationalGTensor = method_not_implemented
    MP2 = method_not_implemented
    CISD = method_not_implemented
    CCSD = method_not_implemented
    CASCI = method_not_implemented
    CASSCF = method_not_implemented


class _CiderDF:
    nuc_grad_method = _CiderKS.nuc_grad_method



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes

import numpy as np

from ciderpress.dft.plans import get_cider_exponent
from ciderpress.lib import load_library as load_cider_library

libcider = load_cider_library("libmcider")


def get_get_exponent(gg_kwargs):
    def _get_exponent(rho_data):
        rho = rho_data[0]
        sigma = np.einsum("xg,xg->g", rho_data[1:4], rho_data[1:4])
        tau = rho_data[4]
        return get_cider_exponent(
            rho,
            sigma,
            tau,
            a0=gg_kwargs["a0"],
            grad_mul=0,
            tau_mul=gg_kwargs["fac_mul"],
            nspin=1,
        )[0]

    return _get_exponent


def get_nonlocal_features(
    rho, coords, vvrho, vvweight, vvcoords, get_exponent_r, get_exponent_rp, version="i"
):
    thresh = 1e-8
    nfeat = 12 if version == "i" else 4
    feat = np.zeros((rho[0].size, nfeat))

    threshind = rho[0] >= thresh
    coords = coords[threshind]
    expnt_data = get_exponent_r(rho[:, threshind])

    innerthreshind = vvrho[0] >= thresh
    vvcoords = vvcoords[innerthreshind]
    vvweight = vvweight[innerthreshind]
    vvexpnt_data = get_exponent_rp(vvrho[:, innerthreshind])

    vvf = vvrho[0, innerthreshind] * vvweight

    vvcoords = np.asarray(vvcoords, order="C")
    coords = np.asarray(coords, order="C")
    _feat = np.empty((expnt_data.shape[0], nfeat))

    if version == "i":
        libcider.debug_numint_vi(
            _feat.ctypes.data_as(ctypes.c_void_p),
            vvexpnt_data.ctypes.data_as(ctypes.c_void_p),
            vvf.ctypes.data_as(ctypes.c_void_p),
            vvcoords.ctypes.data_as(ctypes.c_void_p),
            coords.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(vvcoords.shape[0]),
            ctypes.c_int(coords.shape[0]),
        )
    else:
        if version == "j":
            fn = libcider.debug_numint_vj
        elif version == "k":
            fn = libcider.debug_numint_vk
        else:
            raise ValueError("Version must be i, j, or k")
        fn(
            _feat.ctypes.data_as(ctypes.c_void_p),
            vvexpnt_data.ctypes.data_as(ctypes.c_void_p),
            expnt_data.ctypes.data_as(ctypes.c_void_p),
            vvf.ctypes.data_as(ctypes.c_void_p),
            vvcoords.ctypes.data_as(ctypes.c_void_p),
            coords.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(vvcoords.shape[0]),
            ctypes.c_int(coords.shape[0]),
            ctypes.c_double(2.0),
        )
    feat[threshind, :] = _feat
    return feat



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import os

import numpy as np
from pyscf.lib import chkfile, prange
from scipy.linalg import cho_solve, cholesky

from ciderpress.dft.plans import get_rho_tuple_with_grad_cross, vxc_tuple_to_array
from ciderpress.dft.settings import FeatureSettings
from ciderpress.dft.xc_evaluator import MappedXC
from ciderpress.dft.xc_evaluator2 import MappedXC2


def strk_to_tuplek(d, ref=False):
    if not isinstance(d, dict):
        raise NotImplementedError
    nd = {}
    for occ, vocc in d.items():
        for num, vnum in vocc.items():
            if ref:
                nd[occ, int(num)] = vnum
                continue
            if isinstance(vnum, (tuple, list)):
                s = vnum[0]
                v = vnum[1]
            else:
                s = 0
                v = vnum
            nd[occ, int(num)] = (s, v)
    return nd


class MOLGP:
    """
    Gaussian process model for the exchange-correlation functional
    or its components.
    """

    def __init__(
        self,
        kernels,
        settings,
        libxc_baseline=None,
        default_noise=0.030,
    ):
        """
        Args:
            kernels (list[DFTKernel]): List of kernels that sum to the XC energy.
            settings (DescParams or FeatureSettings): Settings for the
                features. Specifies what the feature vector is.
            libxc_baseline (str or None): Additional baseline functional to
                evaluate using the libxc library.
            default_noise (float): Default noise hyperparameter, used if noise
                is not provided for a particular data point
        """
        self.libxc_baseline = libxc_baseline
        if not isinstance(settings, FeatureSettings):
            raise ValueError("settings must be FeatureSettings")
        self.settings = settings
        if self.libxc_baseline is not None:
            raise NotImplementedError
        if not isinstance(kernels, list):
            kernels = [kernels]
        if len(kernels) < 1:
            raise ValueError("Need at least 1 covariance kernel")
        self.kernels = kernels
        self.default_noise = default_noise

        self.numerical_epsilon = 1e-9
        self.args = None

        self.reset_reactions()
        self.Xctrl_exch = None
        self.Xctrl_corr = None
        self.y = None
        self.K_ = None
        self.Kcov_ = None
        self.alpha_mol_ = None
        self.y_mol_ = None
        self.ks_baseline_dict = {}
        self.exx_ref_dict = {}
        self.dexx_ref_dict = {}
        # Reference exact exchange values with baseline subtracted
        self.exch_ref_dict = {}
        # Vector of exchange covariances with control points
        self.exch_cov_dict = {}
        # Vector of derivative of exchange covariances with control points,
        # taken with respect to orbital occupation
        self.exch_dcov_dict = {}
        # Baseline for XC energy
        self.corr_bas_dict = {}
        # Vector of correlation covariances with control points
        self.corr_cov_dict = {}
        # Vector of derivative of correlation covariances with control points
        self.corr_dcov_dict = {}

    def map(self, mapping_plans):
        """
        Map the MOLGP model to an Evaluator objec that can efficiently
        evaluate the XC energy and which can also be serialized.

        Returns:
            list(callable): Functions that map each
                kernel to a MappedDFTKernel object
        """
        mapped_kernels = [
            kernel.map(mapping_plan)
            for kernel, mapping_plan in zip(self.kernels, mapping_plans)
        ]
        return MappedXC(
            mapped_kernels,
            self.settings,
            libxc_baseline=self.libxc_baseline,
        )

    @property
    def num_kernels(self):
        return len(self.kernels)

    @property
    def xkernels(self):
        return [k for k in self.kernels if k.component == "x"]

    @property
    def ckernels(self):
        return [k for k in self.kernels if k.component != "x"]

    def fit(self, x=None, sigma_min=0.25):
        Kimn_list = []
        Knmimn = 0
        for kernel in self.kernels:
            Kmm = kernel.get_kctrl()
            Kmn = np.stack(kernel.rxn_cov_list).T
            M = Kmm.shape[0]
            mini_noise = self.numerical_epsilon * np.identity(M)
            L_ = cholesky(Kmm + mini_noise, lower=True)
            Kimn = cho_solve((L_, True), Kmn)
            Knmimn += Kmn.T.dot(Kimn)
            Kimn_list.append(Kimn)
        noise_nn = np.array(self.rxn_noise_list)
        noise_nn = noise_nn**2  # get noise covariance from noise std deviation
        if x is not None:
            Knmimn[:] *= x[0] ** 2
            Kimn_list = [x[0] ** 2 * k for k in Kimn_list]
            noise_nn[:] *= sigma_min + x[1] ** 2
        y = np.array(self.rxn_ref_list)
        if len(noise_nn) == 0:
            raise ValueError("Need to set reactions using add_reactions")
        K = Knmimn + np.diag(noise_nn)
        K += self.numerical_epsilon * np.identity(noise_nn.size)
        self.Kcov_ = Knmimn
        self.K_ = K
        L_ = cholesky(K, lower=True)
        alpha_new = cho_solve((L_, True), y)
        self.alpha_mol_ = alpha_new
        self.y_mol_ = y
        for ik, kernel in enumerate(self.kernels):
            kernel.alpha = np.dot(Kimn_list[ik], alpha_new)

    def compute_likelihood(self, x=None, sigma_min=0.25):
        assert self.alpha_mol_ is not None
        if x is None:
            x = np.array([1.0, 1.0])
        y = self.y_mol_
        noise = (sigma_min + x[1] ** 2) * (self.K_ - self.Kcov_)
        Kfull = x[0] ** 2 * self.Kcov_ + noise
        Lfull = cholesky(Kfull, lower=True)
        vec = np.linalg.solve(Lfull, y)
        likelihood = -0.5 * vec.dot(vec)
        # likelihood = -0.5 * np.dot(y, np.linalg.solve(Kfull, y))
        likelihood -= 0.5 * np.linalg.slogdet(Kfull)[1]
        n = y.size
        likelihood -= n / 2 * np.log(2 * np.pi)
        return likelihood

    def optimize_cov_and_noise_(self, refit=True, sigma_min=0.5):
        assert self.alpha_mol_ is not None
        from scipy.optimize import minimize

        def fun(x, *args):
            return -1 * self.compute_likelihood(x, sigma_min=sigma_min)

        res = minimize(fun, np.array([1.0, 1.0]), tol=1e-3)
        print(res.success, res.fun, res.x)
        if refit:
            self.fit(x=res.x, sigma_min=sigma_min)

    def set_control_points(self, X0T_list, reduce=True):
        """
        Transform raw descriptors X and use them as
        control points.

        Args:
            X: Numpy array of raw features
        """
        for kernel in self.kernels:
            kernel.set_control_points(X0T_list, reduce=reduce)

    def _get_normalized_features(self, desc):
        return self.settings.normalizers.get_normalized_feature_vector(desc)

    def _get_normalized_feature_derivs(self, desc, ddesc):
        return self.settings.normalizers.get_derivative_of_normed_features(desc, ddesc)

    @staticmethod
    def load_data(ddir, mol_id, get_orb_deriv):
        fname = os.path.join(ddir["REF"], mol_id + ".hdf5")
        all_data = chkfile.load(fname, "train_data")
        all_data["desc"] = []
        all_data["ddesc"] = []
        for feat_type in ["SL", "NLDF", "NLOF", "SDMX", "HYB"]:
            if ddir[feat_type] is None:
                if feat_type == "SL":
                    raise ValueError("Need semilocal features")
                else:
                    continue
            else:
                fname = os.path.join(ddir[feat_type], mol_id + ".hdf5")
                data = chkfile.load(fname, "train_data")
                all_data["desc"].append(data["desc"])
                if get_orb_deriv or (get_orb_deriv is None and "ddesc" in data):
                    all_data["ddesc"].append(data["ddesc"])
                    get_orb_deriv = True
                else:
                    get_orb_deriv = False
        all_data["desc"] = np.concatenate(all_data["desc"], axis=-2)
        assert all_data["desc"].ndim == 3
        if get_orb_deriv:
            ddesc = all_data["ddesc"]
            all_data["ddesc"] = {}
            ddesc0 = ddesc[0]
            for k1, v1 in ddesc0.items():
                all_data["ddesc"][k1] = {}
                for k2, v2 in ddesc0[k1].items():
                    all_data["ddesc"][k1][k2] = []
                    if all_data["desc"].shape[0] == 2:
                        assert len(v2) == 2
                        spin = ddesc0[k1][k2][0]
                        all_data["ddesc"][k1][k2] = []
                        for sub_ddesc in ddesc:
                            assert spin == sub_ddesc[k1][k2][0]
                            all_data["ddesc"][k1][k2].append(sub_ddesc[k1][k2][1])
                        all_data["ddesc"][k1][k2] = (
                            spin,
                            np.concatenate(all_data["ddesc"][k1][k2], axis=0),
                        )
                        assert all_data["ddesc"][k1][k2][1].ndim == 2
                    else:
                        for sub_ddesc in ddesc:
                            all_data["ddesc"][k1][k2].append(sub_ddesc[k1][k2])
                        all_data["ddesc"][k1][k2] = np.concatenate(
                            all_data["ddesc"][k1][k2], axis=0
                        )
                        assert all_data["ddesc"][k1][k2].ndim == 2
        else:
            all_data.pop("ddesc")
        return all_data

    def _compute_mol_covs(
        self, ddir, mol_ids, kernel, get_orb_deriv=None, save_refs=False
    ):
        # mode should be x or c
        blksize = 10000
        if self.args is not None:
            debug_model = self.args.debug_model
            debug_spline = self.args.debug_spline
        else:
            debug_model = None
            debug_spline = None

        deriv = None
        for i, mol_id in enumerate(mol_ids):
            print("MOL ID", mol_id)
            data = self.load_data(ddir, mol_id, get_orb_deriv)
            if deriv is None:
                if get_orb_deriv is None:
                    if "ddesc" in data:
                        deriv = True
                    else:
                        deriv = False
                else:
                    deriv = get_orb_deriv
            if deriv:
                assert "ddesc" in data

            weights = data["wt"]
            # nspin = data['nspin']
            desc = data["desc"]
            vwrtt_tot = 0
            baseline = 0
            if debug_spline is not None:
                etst = 0
            if deriv:
                ddesc = strk_to_tuplek(data["ddesc"])
                self.dexx_ref_dict[mol_id] = strk_to_tuplek(data["dval"], ref=True)
                dvwrtt_tot = {k: 0 for k in ddesc.keys()}
                dbaseline = {k: 0 for k in ddesc.keys()}
            for i0, i1 in prange(0, weights.size, blksize):
                X0T = self._get_normalized_features(desc[..., i0:i1])
                wt = weights[i0:i1]
                m, dm = kernel.multiplicative_baseline(X0T)
                a, da = kernel.additive_baseline(X0T)
                if debug_spline is not None:
                    tmp, dtmp = debug_spline(X0T, rhocut=1e-7)
                    etst += np.dot(tmp, wt)
                # k is (Nctrl, nspin, Nsamp)
                # dk is (Nctrl, nspin, N0, Nsamp)
                if deriv:
                    k, dkdX0T = kernel.get_k_and_deriv(X0T)
                else:
                    k = kernel.get_k(X0T)
                    dkdX0T = None
                # TODO setting nan to zero could cover up more serious issues,
                # but it is the easiest way to take care of small density
                # training data without editing functions used at eval-time.
                if kernel.mode == "SEP":
                    cond = X0T[:, 0] < 1e-6
                    m[cond] = 0.0
                    a[cond] = 0.0
                    k[:, cond] = 0.0
                    for s in range(X0T.shape[0]):
                        dm[s][:, cond[s]] = 0.0
                        da[s][:, cond[s]] = 0.0
                        if deriv:
                            dkdX0T[:, s][:, :, cond[s]] = 0.0
                else:
                    cond = X0T[:, 0].sum(0) < 1e-6
                    m[..., cond] = 0.0
                    dm[..., cond] = 0.0
                    a[..., cond] = 0.0
                    da[..., cond] = 0.0
                    k[..., cond] = 0.0
                    if deriv:
                        print(dkdX0T.shape)
                        dkdX0T[..., cond] = 0.0
                if kernel.mode == "SEP":
                    km = (k * m).sum(1)
                    if deriv:
                        dkm = dkdX0T * m[:, None, :]
                        dkm += k[:, :, None, :] * dm
                else:
                    km = k * m
                    if deriv:
                        dkm = dkdX0T * m + dm * k[:, None, None, :]
                vwrtt_tot += (km * wt).sum(axis=1)
                baseline += (a * wt).sum()
                if deriv:
                    for orb, (s, ddesc_tmp) in ddesc.items():
                        # ddesc_tmp has shape (N0, Nsamp)
                        # avoid overwriting slice
                        ddesc_tmp = wt * self._get_normalized_feature_derivs(
                            desc[s, :, i0:i1], ddesc_tmp[:, i0:i1]
                        )
                        dvwrtt_tot[orb] += np.einsum("cdn,dn->c", dkm[:, s], ddesc_tmp)
                        dbaseline[orb] += (da[s] * ddesc_tmp).sum()
            kernel.cov_dict[mol_id] = vwrtt_tot
            kernel.base_dict[mol_id] = baseline
            if debug_model is not None:
                diff = debug_model.kernels[0].alpha.dot(vwrtt_tot)
                print("REF VWRTT", diff, kernel.base_dict[mol_id])
            if debug_spline is not None:
                print("DEBUG EN", etst)
            print("BASELINE", baseline)
            if save_refs:
                self.exx_ref_dict[mol_id] = (data["val"] * weights).sum()
                print("exx_ref", self.exx_ref_dict[mol_id])
                self.ks_baseline_dict[mol_id] = data["e_tot_orig"] - data["exc_orig"]
            if deriv:
                kernel.dcov_dict[mol_id] = dvwrtt_tot
                kernel.dbase_dict[mol_id] = dbaseline
                print("DBASELINE", dbaseline)
                if debug_model is not None:
                    diff = {
                        k: debug_model.kernels[0].alpha.dot(dvwrtt)
                        for k, dvwrtt in dvwrtt_tot.items()
                    }
                    print("REF DVWRTT", diff, kernel.dbase_dict[mol_id])

    def store_mol_covs(self, ddir, mol_ids, get_orb_deriv=None, get_correlation=True):
        """
        Reads features (and optionally their derivatives with respect to orbital
        occupations) and computes covariances with the control point sets of all
        the kernels of the models (or only the exchange components if
        get_correlation=False). Also stores the reference energy data
        in the dataset

        Args:
            ddir (str or dict[str]): If using DescParams (old version), should
                be single directory with the feature vectors. If using
                FeatureSettings (new version), should be dictionary of
                directories, with 'REF', 'SL', 'NLDF', 'NLOF', 'SDMX', and
                'HYB' being the keys. All are optional except 'SL' and 'REF'.
                'HYB' stands for reference data. It should contain
                reference energies and raw, spin-polarized
                density/gradient/kinetic energy data.
            mol_ids (list[str]): system ids to search for in ddir
            get_orb_deriv (bool): Whether to read the orbital occupation
                derivatives. If None, reads them iff they are available.
                If True, reads them and throws an error if they are not
                available. If False, does not read them.
            get_correlation (bool): Whether to compute covariance for the
                correlation kernels (True) or just the exchange kernels
                (False).
        """
        for i, kernel in enumerate(self.kernels):
            print(ddir, get_orb_deriv, get_correlation)
            save_refs = i == 0
            if isinstance(get_orb_deriv, (list, tuple)):
                deriv = get_orb_deriv[i]
            else:
                deriv = get_orb_deriv
            if get_correlation or kernel.component == "x":
                self._compute_mol_covs(
                    ddir, mol_ids, kernel, get_orb_deriv=deriv, save_refs=save_refs
                )

    def reset_reactions(self):
        self.rxn_ref_list = []
        self.rxn_noise_list = []
        for kernel in self.kernels:
            kernel.rxn_cov_list = []

    def add_reactions(self, rxn_list):
        """
        Store the reactions in rxn_list in the model. These will be used
        as training points when fit() is called. Note: For every mol_id
        present in the reactions, the corresponding system must have been
        added to the covariance dictionary via the store_mol_covs
        function.

        Args:
            rxn_list: List of tuples. In each tuple, element 0
                is an integer mode (mode: 0 (X), 1 (C), or 2 (XC)),
                and element 1 is a reaction dict. Each dict should have

                * structs: struct codes list
                  (If an element in structs is a tuple, the first
                  tuple element is a structure code, and the second
                  is an orbital code.)
                * counts: count codes list
                * energy: Energy of the reaction
                * unit: Eh per (rxn energy unit)
                * noise (optional): noise in Eh for the reaction
        """
        for mode, rxn in rxn_list:
            if mode == 1:
                raise NotImplementedError
            elif mode > 2:
                raise ValueError("Unsupported mode")
            rxn_ref = 0
            if mode == 0:
                for sysid, count in zip(rxn["structs"], rxn["counts"]):
                    if isinstance(sysid, tuple):
                        rxn_ref += count * self.dexx_ref_dict[sysid[0]][sysid[1]]
                    else:
                        rxn_ref += count * self.exx_ref_dict[sysid]
            for kernel in self.xkernels:
                rxn_cov = 0
                for sysid, count in zip(rxn["structs"], rxn["counts"]):
                    if isinstance(sysid, tuple):
                        rxn_cov += count * kernel.dcov_dict[sysid[0]][sysid[1]]
                        rxn_ref -= count * kernel.dbase_dict[sysid[0]][sysid[1]]
                    else:
                        rxn_cov += count * kernel.cov_dict[sysid]
                        rxn_ref -= count * kernel.base_dict[sysid]
                kernel.rxn_cov_list.append(rxn_cov)
            if mode == 2:
                if rxn.get("unit") is None:
                    rxn["unit"] = 0.00159360109742136  # kcal/mol per Ha
                rxn_ref += rxn["energy"] * rxn["unit"]
                for sysid, count in zip(rxn["structs"], rxn["counts"]):
                    rxn_ref -= count * self.ks_baseline_dict[sysid]
                for kernel in self.ckernels:
                    rxn_cov = 0
                    for sysid, count in zip(rxn["structs"], rxn["counts"]):
                        if isinstance(sysid, tuple):
                            rxn_cov += count * kernel.dcov_dict[sysid[0]][sysid[1]]
                            rxn_ref -= count * kernel.dbase_dict[sysid[0]][sysid[1]]
                        else:
                            rxn_cov += count * kernel.cov_dict[sysid]
                            rxn_ref -= count * kernel.base_dict[sysid]
                    kernel.rxn_cov_list.append(rxn_cov)
            else:
                for kernel in self.ckernels:
                    kernel.rxn_cov_list.append(np.zeros(kernel.Nctrl))
            self.rxn_ref_list.append(rxn_ref)
            if rxn.get("noise") is not None:
                noise = rxn["noise"]
            elif rxn.get("noise_factor") is not None:
                noise = rxn["noise_factor"] * self.default_noise
            else:
                noise = self.default_noise
            if rxn.get("noise_rel_factor") is not None:
                noise += rxn["noise_rel_factor"] * np.abs(rxn_ref)
            if rxn.get("weight") is not None:
                noise /= np.sqrt(rxn["weight"])
            self.rxn_noise_list.append(noise)


class MOLGP2(MOLGP):
    """
    Gaussian process model for the exchange-correlation functional
    or its components.
    """

    def __init__(
        self,
        kernels,
        settings,
        libxc_baseline=None,
        default_noise=0.030,
    ):
        """
        Same as MOLGP, except kernels should be of type DFTKernel2.
        This requires a new class because DFTKernel2 has a different approach
        to evaluating baseline functional contributions, so a few
        functions in MOLGP need to be modified.

        Args:
            kernels (list[DFTKernel2]): List of kernels that sum to the XC energy.
            settings (DescParams or FeatureSettings): Settings for the
                features. Specifies what the feature vector is.
            libxc_baseline (str or None): Additional baseline functional to
                evaluate using the libxc library.
            default_noise (float): Default noise hyperparameter, used if noise
                is not provided for a particular data point
        """
        super(MOLGP2, self).__init__(
            kernels,
            settings,
            libxc_baseline=libxc_baseline,
            default_noise=default_noise,
        )

    def _compute_mol_covs(
        self, ddir, mol_ids, kernel, get_orb_deriv=None, save_refs=False
    ):
        # need to have different _compute_mol_covs function to handle baselines differently.
        blksize = 10000

        deriv = None
        for mol_id in mol_ids:
            print("MOL ID", mol_id)
            data = self.load_data(ddir, mol_id, get_orb_deriv)
            if deriv is None:
                if get_orb_deriv is None:
                    if "ddesc" in data:
                        deriv = True
                    else:
                        deriv = False
                else:
                    deriv = get_orb_deriv
            if deriv:
                assert "ddesc" in data
                assert "drho_data" in data

            weights = data["wt"]
            nspin = data["nspin"]
            desc = data["desc"]
            rho_data = data["rho_data"] / nspin
            vwrtt_tot = 0
            baseline = 0
            is_mgga = self.settings.sl_settings.level
            if deriv:
                ddesc = strk_to_tuplek(data["ddesc"])
                drho_data = strk_to_tuplek(data["drho_data"])
                for orb, (s, drho) in drho_data.items():
                    drho[:] /= nspin
                self.dexx_ref_dict[mol_id] = strk_to_tuplek(data["dval"], ref=True)
                dvwrtt_tot = {k: 0 for k in ddesc.keys()}
                dbaseline = {k: 0 for k in ddesc.keys()}
            for i0, i1 in prange(0, weights.size, blksize):
                X0T = self._get_normalized_features(desc[..., i0:i1])
                rho_tuple = get_rho_tuple_with_grad_cross(
                    rho_data[..., i0:i1], is_mgga=is_mgga
                )
                wt = weights[i0:i1]
                m_res = kernel.multiplicative_baseline(rho_tuple)
                a_res = kernel.additive_baseline(rho_tuple)
                dm = vxc_tuple_to_array(rho_data[..., i0:i1], m_res[1:])
                m = m_res[0]
                if a_res is None:
                    da = np.zeros_like(dm)
                    a = np.zeros_like(m)
                else:
                    da = vxc_tuple_to_array(rho_data[..., i0:i1], a_res[1:])
                    a = a_res[0]

                # k is (Nctrl, nspin, Nsamp)
                # dk is (Nctrl, nspin, N0, Nsamp)
                if deriv:
                    k, dkdX0T = kernel.get_k_and_deriv(X0T)
                else:
                    k = kernel.get_k(X0T)
                    dkdX0T = None
                # TODO setting nan to zero could cover up more serious issues,
                # but it is the easiest way to take care of small density
                # training data without editing functions used at eval-time.
                cond = rho_tuple[0] < 1e-6
                if kernel.mode == "SEP":
                    m[cond] = 0.0
                    a[cond] = 0.0
                    k[:, cond] = 0.0
                    if deriv:
                        for s in range(X0T.shape[0]):
                            dm[s][:, cond[s]] = 0.0
                            da[s][:, cond[s]] = 0.0
                            dkdX0T[:, s][:, :, cond[s]] = 0.0
                else:
                    if cond.shape[0] == 1:
                        scond = cond[0]
                    else:
                        scond = np.logical_and(cond[0], cond[1])
                    m[..., scond] = 0.0
                    a[..., scond] = 0.0
                    k[..., scond] = 0.0
                    if deriv:
                        for s in range(X0T.shape[0]):
                            dkdX0T[:, s, cond[s], :] = 0.0
                            dm[s][:, cond[s]] = 0.0
                            da[s][:, cond[s]] = 0.0
                if kernel.mode == "SEP":
                    km = (k * m).sum(1)
                    if deriv:
                        dkm1 = dkdX0T * m[:, None, :]
                        dkm2 = k[:, :, None, :] * dm
                else:
                    km = k * m
                    if deriv:
                        dkm1 = dkdX0T * m
                        dkm2 = k[:, None, None, :] * dm
                vwrtt_tot += (km * wt).sum(axis=1)
                baseline += (a * wt).sum()
                if deriv:
                    for orb, (s, ddesc_tmp) in ddesc.items():
                        # ddesc_tmp has shape (N0, Nsamp)
                        # avoid overwriting slice
                        ddesc_tmp = wt * self._get_normalized_feature_derivs(
                            desc[s, :, i0:i1], ddesc_tmp[:, i0:i1]
                        )
                        drho_tmp = wt * drho_data[orb][:, i0:i1]
                        dvwrtt_tot[orb] += np.einsum("cdn,dn->c", dkm1[:, s], ddesc_tmp)
                        dvwrtt_tot[orb] += np.einsum("cdn,dn->c", dkm2[:, s], drho_tmp)
                        dbaseline[orb] += np.dot(da * drho_tmp, wt)
            kernel.cov_dict[mol_id] = vwrtt_tot
            kernel.base_dict[mol_id] = baseline
            if save_refs:
                self.exx_ref_dict[mol_id] = (data["val"] * weights).sum()
                self.ks_baseline_dict[mol_id] = data["e_tot_orig"] - data["exc_orig"]
            if deriv:
                kernel.dcov_dict[mol_id] = dvwrtt_tot
                kernel.dbase_dict[mol_id] = dbaseline

    def map(self, mapping_plans):
        """
        Map the MOLGP model to an Evaluator objec that can efficiently
        evaluate the XC energy and which can also be serialized.

        Returns:
            list(callable): Functions that map each
                kernel to a MappedDFTKernel object
        """
        mapped_kernels = [
            kernel.map(mapping_plan)
            for kernel, mapping_plan in zip(self.kernels, mapping_plans)
        ]
        return MappedXC2(
            mapped_kernels,
            self.settings,
            libxc_baseline=self.libxc_baseline,
        )



import numpy as np
import pyscf.df.grad.rks as rks_grad_df
from pyscf import lib
from pyscf.dft import gen_grid
from pyscf.grad import rks as rks_grad
from pyscf.grad.rks import (
    _gga_grad_sum_,
    _tau_grad_dot_,
    get_nlc_vxc,
    get_nlc_vxc_full_response,
    grids_noresponse_cc,
    grids_response_cc,
)
from pyscf.lib import logger


def get_veff(ks_grad, mol=None, dm=None):
    if mol is None:
        mol = ks_grad.mol
    if dm is None:
        dm = ks_grad.base.make_rdm1()
    t0 = (logger.process_clock(), logger.perf_counter())

    mf = ks_grad.base
    ni = mf._numint
    grids, nlcgrids = rks_grad._initialize_grids(ks_grad)

    # NOTE pyscf has hyb params here, shouldn't be needed for now

    mem_now = lib.current_memory()[0]
    max_memory = max(2000, ks_grad.max_memory * 0.9 - mem_now)
    if ks_grad.grid_response:
        exc, vxc = get_vxc_full_response(
            ni, mol, grids, mf.xc, dm, max_memory=max_memory, verbose=ks_grad.verbose
        )
        if mf.do_nlc():
            if ni.libxc.is_nlc(mf.xc):
                xc = mf.xc
            else:
                xc = mf.nlc
            enlc, vnlc = get_nlc_vxc_full_response(
                ni,
                mol,
                nlcgrids,
                xc,
                dm,
                max_memory=max_memory,
                verbose=ks_grad.verbose,
            )
            exc += enlc
            vxc += vnlc
        logger.debug1(ks_grad, "sum(grids response) %s", exc.sum(axis=0))
    else:
        exc, vxc = get_vxc(
            ni, mol, grids, mf.xc, dm, max_memory=max_memory, verbose=ks_grad.verbose
        )
        if mf.do_nlc():
            if ni.libxc.is_nlc(mf.xc):
                xc = mf.xc
            else:
                xc = mf.nlc
            enlc, vnlc = get_nlc_vxc(
                ni,
                mol,
                nlcgrids,
                xc,
                dm,
                max_memory=max_memory,
                verbose=ks_grad.verbose,
            )
            vxc += vnlc
    t0 = logger.timer(ks_grad, "vxc", *t0)

    has_df = hasattr(mf, "with_df") and mf.with_df is not None
    if not has_df:
        if not ni.libxc.is_hybrid_xc(mf.xc):
            vj = ks_grad.get_j(mol, dm)
            vxc += vj
        else:
            raise NotImplementedError("Hybrid DFT with CIDER functionals")
    else:
        if not ni.libxc.is_hybrid_xc(mf.xc):
            vj = ks_grad.get_j(mol, dm)
            vxc += vj
            if ks_grad.auxbasis_response:
                e1_aux = vj.aux.sum((0, 1))
        else:
            raise NotImplementedError("Hybrid DFT with CIDER functionals")

        if ks_grad.auxbasis_response:
            logger.debug1(ks_grad, "sum(auxbasis response) %s", e1_aux.sum(axis=0))
            vxc = lib.tag_array(vxc, exc1_grid=exc, aux=e1_aux)
        else:
            vxc = lib.tag_array(vxc, exc1_grid=exc)
        return vxc

    return lib.tag_array(vxc, exc1_grid=exc)


def get_vxc(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    if not ni.settings.nldf_settings.is_empty:
        return get_vxc_nldf(
            ni,
            mol,
            grids,
            xc_code,
            dms,
            relativity=relativity,
            hermi=hermi,
            max_memory=max_memory,
            verbose=verbose,
        )
    ni.timer.start("nr_rks")
    if relativity != 0:
        raise NotImplementedError

    xctype = ni.settings.sl_settings.level
    ni.initialize_feature_generators(mol, grids, 1)

    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_loc = mol.ao_loc_nr()

    if dms.ndim == 2:
        dms = dms[None, :, :]

    nelec = np.zeros(nset)
    excsum = np.zeros(nset)

    ao_deriv = 2
    vmat = np.zeros((nset, 3, nao, nao))

    def block_loop(ao_deriv):
        if ni.has_sdmx:
            raise NotImplementedError
        else:
            extra_ao = 0
        for ao, mask, weight, coords in ni.block_loop(
            mol, grids, nao, ao_deriv, max_memory=max_memory, extra_ao=extra_ao
        ):
            for i in range(nset):
                rho = make_rho(i, ao, mask, xctype)
                ni.timer.start("xc cider")
                exc, (vxc, vxc_nldf, vxc_sdmx) = ni.eval_xc_cider(
                    xc_code, rho, None, None, deriv=1, xctype=xctype
                )[:2]
                ni.timer.stop("xc cider")
                assert vxc_nldf is None
                den = rho[0] * weight
                nelec[i] += den.sum()
                excsum[i] += np.dot(den, exc)
                wv = weight * vxc
                yield i, ao, mask, wv

    for idm, ao, mask, wv in block_loop(ao_deriv):
        wv[0] *= 0.5
        _gga_grad_sum_(vmat[idm], mol, ao, wv, mask, ao_loc)
        if xctype == "MGGA":
            wv[4] *= 0.5
            _tau_grad_dot_(vmat[idm], mol, ao, wv[4], mask, ao_loc, True)

    exc = None
    if nset == 1:
        vmat = vmat[0]
    # - sign because nabla_X = -nabla_x
    return exc, -vmat


def get_vxc_nldf(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    if not hasattr(grids, "grids_indexer"):
        raise ValueError("Grids object must have indexer for NLDF evaluation")

    if relativity != 0:
        raise NotImplementedError

    xctype = ni.settings.sl_settings.level
    ni.initialize_feature_generators(mol, grids, 1)
    assert ni.has_nldf

    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_loc = mol.ao_loc_nr()

    if dms.ndim == 2:
        dms = dms[None, :, :]

    nelec = np.zeros(nset)
    excsum = np.zeros(nset)
    ao_deriv = 2
    vmat = np.zeros((nset, 3, nao, nao))
    nrho = 5 if xctype == "MGGA" else 4
    if not ni.settings.nlof_settings.is_empty:
        raise NotImplementedError
    rho_full = np.zeros((nset, nrho, grids.weights.size), dtype=np.float64, order="C")
    wv_full = np.zeros((nset, nrho, grids.weights.size), dtype=np.float64, order="C")

    num_nldf = ni.settings.nldf_settings.nfeat
    vxc_nldf_full = np.zeros(
        (nset, num_nldf, grids.weights.size), dtype=np.float64, order="C"
    )

    def block_loop(ao_deriv):
        ip0 = 0
        for ao, mask, weight, coords in ni.block_loop(
            mol, grids, nao, ao_deriv, max_memory=max_memory
        ):
            ip1 = ip0 + weight.size
            for i in range(nset):
                yield i, ip0, ip1, ao, mask, weight, coords
            ip0 = ip1

    if any(x in xc_code.upper() for x in ("CC06", "CS", "BR89", "MK00")):
        raise NotImplementedError("laplacian in meta-GGA method")
    for i, ip0, ip1, ao, mask, weight, coords in block_loop(ao_deriv):
        rho_full[i, :, ip0:ip1] = make_rho(i, ao, mask, xctype)

    par_atom = False
    extra_ao = ni.nldfgen.get_extra_ao()
    if ni.has_sdmx:
        raise NotImplementedError
    if par_atom:
        raise NotImplementedError
    else:
        nldf_feat = []
        for idm in range(nset):
            nldf_feat.append(ni.nldfgen.get_features(rho_full[idm]))
        ip0 = 0
        for mask, weight, coords in ni.extra_block_loop(
            mol, grids, max_memory=max_memory, extra_ao=extra_ao
        ):
            ip1 = ip0 + weight.size
            for idm in range(nset):
                rho = rho_full[idm, :, ip0:ip1]
                sdmx_feat = None
                exc, (vxc, vxc_nldf, vxc_sdmx) = ni.eval_xc_cider(
                    xc_code,
                    np.ascontiguousarray(rho),
                    np.ascontiguousarray(nldf_feat[idm][:, ip0:ip1]),
                    sdmx_feat,
                    deriv=1,
                    xctype=xctype,
                )[:2]
                vxc_nldf_full[idm, :, ip0:ip1] = vxc_nldf * weight
                den = rho[0] * weight
                nelec[idm] += den.sum()
                excsum[idm] += np.dot(den, exc)
                wv_full[idm, :, ip0:ip1] = weight * vxc
            ip0 = ip1
        for idm in range(nset):
            wv_full[idm, :, :] += ni.nldfgen.get_potential(vxc_nldf_full[idm])

    for i, ip0, ip1, ao, mask, weight, coords in block_loop(ao_deriv):
        for idm in range(nset):
            rho = np.ascontiguousarray(rho_full[idm, :, ip0:ip1])
            wv = np.ascontiguousarray(wv_full[idm][:, ip0:ip1])
            wv[0] *= 0.5
            _gga_grad_sum_(vmat[idm], mol, ao, wv[:4], mask, ao_loc)
            if xctype == "MGGA":
                wv[4] *= 0.5
                _tau_grad_dot_(vmat[idm], mol, ao, wv[4], mask, ao_loc, True)

    exc = None
    if nset == 1:
        vmat = vmat[0]
    # - sign because nabla_X = -nabla_x
    return exc, -vmat


def get_vxc_full_response(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    if not ni.settings.nldf_settings.is_empty:
        return get_vxc_nldf_full_response(
            ni,
            mol,
            grids,
            xc_code,
            dms,
            relativity=relativity,
            hermi=hermi,
            max_memory=max_memory,
            verbose=verbose,
        )

    if not ni.settings.sdmx_settings.is_empty:
        raise NotImplementedError
    if not ni.settings.nlof_settings.is_empty:
        raise NotImplementedError
    xctype = ni.settings.sl_settings.level
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_loc = mol.ao_loc_nr()

    excsum = np.zeros((mol.natm, 3))
    vmat = np.zeros((3, nao, nao))

    ao_deriv = 2
    for atm_id, (coords, weight, weight1) in enumerate(grids_response_cc(grids)):
        mask = gen_grid.make_mask(mol, coords)
        ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask, cutoff=grids.cutoff)
        rho = make_rho(0, ao[:10], mask, xctype)
        exc, (vxc, vxc_nldf, vxc_sdmx) = ni.eval_xc_cider(
            xc_code,
            rho,
            None,
            None,
            deriv=1,
            xctype=xctype,
        )[:2]
        wv = weight * vxc
        wv[0] *= 0.5
        vtmp = np.zeros((3, nao, nao))
        _gga_grad_sum_(vtmp, mol, ao, wv, mask, ao_loc)
        if xctype == "MGGA":
            wv[4] *= 0.5  # for the factor 1/2 in tau
            _tau_grad_dot_(vtmp, mol, ao, wv[4], mask, ao_loc, True)
        vmat += vtmp
        # response of weights
        excsum += np.einsum("r,r,nxr->nx", exc, rho[0], weight1)
        # response of grids coordinates
        excsum[atm_id] += np.einsum("xij,ji->x", vtmp, dms) * 2
        rho = vxc = wv = None
    # - sign because nabla_X = -nabla_x
    return excsum, -vmat


def get_vxc_nldf_full_response(
    ni, mol, grids, xc_code, dms, relativity=0, hermi=1, max_memory=2000, verbose=None
):
    if not hasattr(grids, "grids_indexer"):
        raise ValueError("Grids object must have indexer for NLDF evaluation")

    if relativity != 0:
        raise NotImplementedError

    xctype = ni.settings.sl_settings.level
    assert ni.has_nldf
    ni.initialize_feature_generators(mol, grids, 1)

    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    if nset != 1:
        raise NotImplementedError
    ao_loc = mol.ao_loc_nr()

    assert dms.ndim == 2

    nelec = np.zeros(nset)
    vmat = np.zeros((3, nao, nao))
    nrho = 5 if xctype == "MGGA" else 4
    if not ni.settings.nlof_settings.is_empty:
        raise NotImplementedError

    num_nldf = ni.settings.nldf_settings.nfeat
    vxc_nldf_full = np.zeros(
        (nset, num_nldf, grids.weights.size), dtype=np.float64, order="C"
    )

    ga_loc = ni.nldfgen.interpolator.grid_loc_atom
    assert ga_loc is not None
    rho_full = np.zeros((nset, nrho, ga_loc[-1]), dtype=np.float64, order="C")
    wv_full = np.zeros((nset, nrho, ga_loc[-1]), dtype=np.float64, order="C")
    exc = np.zeros(ga_loc[-1], dtype=np.float64, order="C")
    all_coords = np.zeros((ga_loc[-1], 3), dtype=np.float64, order="C")

    ao_deriv = 1
    for atm_id, (coords, weight) in enumerate(grids_noresponse_cc(grids)):
        ip0, ip1 = ga_loc[atm_id : atm_id + 2]
        mask = gen_grid.make_mask(mol, coords)
        ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask, cutoff=grids.cutoff)
        rho_full[0, :, ip0:ip1] = make_rho(0, ao, mask, xctype)
        all_coords[ip0:ip1] = coords

    par_atom = False
    extra_ao = ni.nldfgen.get_extra_ao()
    if ni.has_sdmx:
        raise NotImplementedError
    if par_atom:
        raise NotImplementedError
    else:
        nldf_feat = [
            ni.nldfgen.get_features(rho_full[0], map_grids=False, grad_mode=True)
        ]
        ip0 = 0
        for mask, weight, coords in ni.extra_block_loop(
            mol, grids, max_memory=max_memory, extra_ao=extra_ao
        ):
            ip1 = ip0 + weight.size
            ip1 = min(ip1, grids.grids_indexer.idx_map.size)
            weight = weight[: ip1 - ip0]
            inds = grids.grids_indexer.idx_map[ip0:ip1]
            rho = rho_full[0][..., inds]
            exc[inds], (vxc, vxc_nldf, vxc_sdmx) = ni.eval_xc_cider(
                xc_code,
                rho,
                nldf_feat[0][:, ip0:ip1],
                None,
                deriv=1,
                xctype=xctype,
            )[:2]
            vxc_nldf_full[0][:, ip0:ip1] = vxc_nldf * weight
            den = rho[0] * weight
            nelec[0] += den.sum()
            wv_full[0][:, inds] = weight * vxc
            ip0 = ip1
        wvtmp, cidergg_g, excsum = ni.nldfgen.get_potential(
            vxc_nldf_full[0], map_grids=False, grad_mode=True
        )
        # excsum[:] = 0
        wv_full[0] += wvtmp

    exc_ref = 0
    ao_deriv = 2
    for atm_id, (coords, weight, weight1) in enumerate(grids_response_cc(grids)):
        mask = gen_grid.make_mask(mol, coords)
        ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask, cutoff=grids.cutoff)
        ip0, ip1 = ga_loc[atm_id : atm_id + 2]
        rho = np.ascontiguousarray(rho_full[0, :, ip0:ip1])
        wv = np.ascontiguousarray(wv_full[0][:, ip0:ip1])
        vtmp = np.zeros((3, nao, nao))
        wv[0] *= 0.5
        _gga_grad_sum_(vtmp, mol, ao, wv[:4], mask, ao_loc)
        if xctype == "MGGA":
            wv[4] *= 0.5
            _tau_grad_dot_(vtmp, mol, ao, wv[4], mask, ao_loc, True)
        vmat += vtmp

        # response of weights
        excsum += np.einsum("r,r,nxr->nx", exc[ip0:ip1], rho[0], weight1)
        # response of grids coordinates
        excsum[atm_id] += np.einsum("xij,ji->x", vtmp, dms) * 2
        excsum += np.dot(weight1, cidergg_g[ip0:ip1])
        exc_ref += np.sum(exc[ip0:ip1] * rho[0] * weight)
        ip0 = ip1
    # - sign because nabla_X = -nabla_x
    return excsum, -vmat


class Gradients(rks_grad.Gradients):

    get_veff = get_veff


class DFGradients(rks_grad_df.Gradients):

    get_veff = get_veff



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import numpy as np
from gpaw.atom.radialgd import AERadialGridDescriptor
from gpaw.sphere.lebedev import R_nv, Y_nL, weight_n
from gpaw.xc.gga import GGA, radial_gga_vars
from gpaw.xc.mgga import MGGA
from gpaw.xc.pawcorrection import rnablaY_nLv
from numpy import pi, sqrt
from scipy.interpolate import interp1d

from ciderpress.gpaw.gpaw_grids import GCRadialGridDescriptor


def create_kinetic_diffpaw(xcc, ny, phi_jg, tau_ypg, _interpc, r_g_new):
    nj = len(phi_jg)
    dphidr_jg = np.zeros(np.shape(phi_jg))
    for j in range(nj):
        phi_g = phi_jg[j]
        xcc.rgd.derivative(phi_g, dphidr_jg[j])

    phi_jg = _interpc(phi_jg)
    dphidr_jg = _interpc(dphidr_jg)

    # second term
    for y in range(ny):
        i1 = 0
        p = 0
        Y_L = xcc.Y_nL[y]
        for j1, l1, L1 in xcc.jlL:
            for j2, l2, L2 in xcc.jlL[i1:]:
                c = Y_L[L1] * Y_L[L2]
                temp = c * dphidr_jg[j1] * dphidr_jg[j2]
                tau_ypg[y, p, :] += temp
                p += 1
            i1 += 1
    # first term
    for y in range(ny):
        i1 = 0
        p = 0
        rnablaY_Lv = xcc.rnablaY_nLv[y, : xcc.Lmax]
        Ax_L = rnablaY_Lv[:, 0]
        Ay_L = rnablaY_Lv[:, 1]
        Az_L = rnablaY_Lv[:, 2]
        for j1, l1, L1 in xcc.jlL:
            for j2, l2, L2 in xcc.jlL[i1:]:
                temp = Ax_L[L1] * Ax_L[L2] + Ay_L[L1] * Ay_L[L2] + Az_L[L1] * Az_L[L2]
                temp *= phi_jg[j1] * phi_jg[j2]
                # temp[1:] /= r_g_new[1:]**2
                # temp[0] = temp[1]
                temp /= r_g_new**2
                if r_g_new[0] == 0:
                    temp[0] = temp[1]
                tau_ypg[y, p, :] += temp
                p += 1
            i1 += 1
    tau_ypg *= 0.5


class DiffPAWXCCorrection:
    def __init__(
        self,
        rgd,
        nc_g,
        nct_g,
        nc_corehole_g,
        dc_g,
        dct_g,
        dc_corehole_g,
        B_pqL,
        n_qg,
        nt_qg,
        d_qg,
        dt_qg,
        e_xc0,
        Lmax,
        tau_npg=None,
        taut_npg=None,
        tauc_g=None,
        tauct_g=None,
        ke_order_ng=True,
    ):
        self.rgd = rgd
        self.nc_g = nc_g
        self.nct_g = nct_g
        self.dc_g = dc_g
        self.dct_g = dct_g
        self.nc_corehole_g = nc_corehole_g
        self.dc_corehole_g = dc_corehole_g
        self.B_pqL = B_pqL
        self.n_qg = n_qg
        self.nt_qg = nt_qg
        self.d_qg = d_qg
        self.dt_qg = dt_qg
        self.e_xc0 = e_xc0
        self.Lmax = Lmax
        self.tau_npg = tau_npg
        self.taut_npg = taut_npg
        self.tauc_g = tauc_g
        self.tauct_g = tauct_g
        self.Y_nL = Y_nL
        if self.tau_npg is not None:
            NP = self.tau_npg.shape[1]
            if ke_order_ng:
                # radial coordinate is last index
                self.tau_pg = np.ascontiguousarray(
                    self.tau_npg.transpose(1, 0, 2).reshape(NP, -1)
                )
                self.taut_pg = np.ascontiguousarray(
                    self.taut_npg.transpose(1, 0, 2).reshape(NP, -1)
                )
            else:
                # angular coordinate is last index
                self.tau_pg = np.ascontiguousarray(
                    self.tau_npg.transpose(1, 2, 0).reshape(NP, -1)
                )
                self.taut_pg = np.ascontiguousarray(
                    self.taut_npg.transpose(1, 2, 0).reshape(NP, -1)
                )

    @classmethod
    def from_setup(cls, setup, build_kinetic=False, ke_order_ng=True):

        if hasattr(setup, "old_xc_correction"):
            xcc = setup.old_xc_correction
        else:
            setup.old_xc_correction = setup.xc_correction
            xcc = setup.xc_correction

        tab = np.array((2, 10, 18, 36, 54, 86, 118))
        (setup.Z > tab).sum()
        rcut = xcc.rgd.r_g[-1]
        if setup.Z > 36 and hasattr(setup.rgd, "a") and hasattr(setup.rgd, "b"):
            rgd = AERadialGridDescriptor(setup.rgd.a, setup.rgd.b, setup.rgd.N)
            gcut = rgd.ceil(rcut)
            rgd = rgd.new(gcut)
        else:
            rgd = GCRadialGridDescriptor(int(setup.Z), 200)
            gcut = rgd.ceil(rcut)
            rgd = rgd.make_cut(gcut)

        def _interpc(func):
            return interp1d(
                np.arange(xcc.rgd.r_g.size),
                func,
                kind="cubic",
            )(xcc.rgd.r2g(rgd.r_g))

        core_dens = {}
        names = ["nc_g", "nct_g", "nc_corehole_g"]
        n_g_list = [xcc.nc_g, xcc.nct_g, xcc.nc_corehole_g]
        if True:
            names += ["tauc_g", "tauct_g"]
            n_g_list += [xcc.tauc_g, xcc.tauct_g]
        for name, n_g in zip(names, n_g_list):
            if n_g is None:
                continue
            core_dens[name] = n_g
            core_dens["d" + name] = xcc.rgd.derivative(n_g)

        d_qg = [xcc.rgd.derivative(n_g) for n_g in xcc.n_qg]
        dt_qg = [xcc.rgd.derivative(n_g) for n_g in xcc.nt_qg]

        if build_kinetic:
            nii = xcc.nii
            nn = len(xcc.rnablaY_nLv)
            ng = rgd.r_g.shape[0]
            tau_npg = np.zeros((nn, nii, ng))
            taut_npg = np.zeros((nn, nii, ng))
            create_kinetic_diffpaw(xcc, nn, xcc.phi_jg, tau_npg, _interpc, rgd.r_g)
            create_kinetic_diffpaw(xcc, nn, xcc.phit_jg, taut_npg, _interpc, rgd.r_g)
        else:
            tau_npg = None
            taut_npg = None

        return cls(
            rgd,
            _interpc(core_dens["nc_g"]),
            _interpc(core_dens["nct_g"]),
            _interpc(core_dens["nc_corehole_g"])
            if xcc.nc_corehole_g is not None
            else None,
            _interpc(core_dens["dnc_g"]),
            _interpc(core_dens["dnct_g"]),
            _interpc(core_dens["dnc_corehole_g"])
            if xcc.nc_corehole_g is not None
            else None,
            xcc.B_pqL,
            np.array([_interpc(n_g) for n_g in xcc.n_qg]),
            np.array([_interpc(nt_g) for nt_g in xcc.nt_qg]),
            np.array([_interpc(n_g) for n_g in d_qg]),
            np.array([_interpc(nt_g) for nt_g in dt_qg]),
            xcc.e_xc0,
            xcc.Lmax,
            tau_npg,
            taut_npg,
            _interpc(core_dens["tauc_g"]),
            _interpc(core_dens["tauct_g"]),
            ke_order_ng=ke_order_ng,
        )


"""
def create_grad_and_kinetic_diffpaw(xcc, jl, ny, phi_jg, d_qg, tau_ypg, r_g_new):
    nj = len(phi_jg)
    dphidr_jg = np.zeros(np.shape(phi_jg))
    for j in range(nj):
        phi_g = phi_jg[j]
        xcc.rgd.derivative(phi_g, dphidr_jg[j])

    # grad term
    q = 0
    for j1, l1 in jl:
        for j2, l2 in jl[j1:]:
            d_qg[q] = phi_jg[j1] * dphidr_jg[j2]
            d_qg[q] += dphidr_jg[j1] * phi_jg[j2]
            q += 1
    # second term
    for y in range(ny):
        i1 = 0
        p = 0
        Y_L = xcc.Y_nL[y]
        for j1, l1, L1 in xcc.jlL:
            for j2, l2, L2 in xcc.jlL[i1:]:
                c = Y_L[L1] * Y_L[L2]
                temp = c * dphidr_jg[j1] * dphidr_jg[j2]
                tau_ypg[y, p, :] += temp
                p += 1
            i1 += 1
    # first term
    for y in range(ny):
        i1 = 0
        p = 0
        rnablaY_Lv = xcc.rnablaY_nLv[y, :xcc.Lmax]
        Ax_L = rnablaY_Lv[:, 0]
        Ay_L = rnablaY_Lv[:, 1]
        Az_L = rnablaY_Lv[:, 2]
        for j1, l1, L1 in xcc.jlL:
            for j2, l2, L2 in xcc.jlL[i1:]:
                temp = (Ax_L[L1] * Ax_L[L2] + Ay_L[L1] * Ay_L[L2] +
                        Az_L[L1] * Az_L[L2])
                temp *= phi_jg[j1] * phi_jg[j2]
                #temp[1:] /= r_g_new[1:]**2
                #temp[0] = temp[1]
                temp /= r_g_new**2
                if r_g_new[0] == 0:
                    temp[0] = temp[1]
                tau_ypg[y, p, :] += temp
                p += 1
            i1 += 1
    tau_ypg *= 0.5


class DiffPAWXCCorrection:

    def __init__(
                self, rgd, nc_g, nct_g, nc_corehole_g,
                dc_g, dct_g, dc_corehole_g,
                B_pqL, n_qg, nt_qg,
                d_qg, dt_qg, e_xc0,
                tau_npg=None, taut_npg=None,
                tauc_g=None, tauct_g=None,
            ):
        self.rgd = rgd
        self.nc_g = nc_g
        self.nct_g = nct_g
        self.dc_g = dc_g
        self.dct_g = dct_g
        self.nc_corehole_g = nc_corehole_g
        self.dc_corehole_g = dc_corehole_g
        self.B_pqL = B_pqL
        self.n_qg = n_qg
        self.nt_qg = nt_qg
        self.d_qg = d_qg
        self.dt_qg = dt_qg
        self.e_xc0 = e_xc0
        self.tau_npg = tau_npg
        self.taut_npg = taut_npg
        self.tauc_g = tauc_g
        self.tauct_g = tauct_g
        # print('SETUP MINS AE', np.min(self.tauc_g - self.dc_g**2 / (8 * self.nc_g + 1e-16)))
        # print('SETUP MINS PS', np.min(self.tauct_g - self.dct_g**2 / (8 * self.nct_g + 1e-16)))
        if self.tau_npg is not None:
            NP = self.tau_npg.shape[1]
            self.tau_pg = np.ascontiguousarray(
                self.tau_npg.transpose(1,0,2).reshape(NP, -1)
            )
            self.taut_pg = np.ascontiguousarray(
                self.taut_npg.transpose(1,0,2).reshape(NP, -1)
            )

    @classmethod
    def from_setup(cls, setup, build_kinetic=False):

        if hasattr(setup, 'old_xc_correction'):
            xcc = setup.old_xc_correction
        else:
            setup.old_xc_correction = setup.xc_correction
            xcc = setup.xc_correction

        rgd = xcc.rgd

        core_dens = {}
        names = ['nc_g', 'nct_g', 'nc_corehole_g']
        n_g_list = [xcc.nc_g, xcc.nct_g, xcc.nc_corehole_g]
        if True:
            names += ['tauc_g', 'tauct_g']
            n_g_list += [xcc.tauc_g, xcc.tauct_g]
        for name, n_g in zip(names, n_g_list):
            if n_g is None:
                continue
            core_dens[name] = n_g
            core_dens['d' + name] = xcc.rgd.derivative(n_g)

        #d_qg = [xcc.rgd.derivative(n_g) for n_g in xcc.n_qg]
        #dt_qg = [xcc.rgd.derivative(n_g) for n_g in xcc.nt_qg]

        nii = xcc.nii
        nn = len(xcc.rnablaY_nLv)
        ng = rgd.r_g.shape[0]
        nq = len(xcc.n_qg)
        d_qg = np.zeros((nq, ng))
        dt_qg = np.zeros((nq, ng))
        tau_npg = np.zeros((nn, nii, ng))
        taut_npg = np.zeros((nn, nii, ng))
        jl = list(enumerate(setup.l_j))
        create_grad_and_kinetic_diffpaw(
            xcc, jl, nn, xcc.phi_jg, d_qg, tau_npg, rgd.r_g
        )
        create_grad_and_kinetic_diffpaw(
            xcc, jl, nn, xcc.phit_jg, dt_qg, taut_npg, rgd.r_g
        )

        return cls(
            rgd,
            core_dens['nc_g'],
            core_dens['nct_g'],
            core_dens['nc_corehole_g'] if xcc.nc_corehole_g is not None else None,
            core_dens['dnc_g'],
            core_dens['dnct_g'],
            core_dens['dnc_corehole_g'] if xcc.nc_corehole_g is not None else None,
            xcc.B_pqL,
            np.asarray([n_g for n_g in xcc.n_qg], order='C'),
            np.asarray([nt_g for nt_g in xcc.nt_qg], order='C'),
            np.asarray([n_g for n_g in d_qg], order='C'),
            np.asarray([nt_g for nt_g in dt_qg], order='C'),
            xcc.e_xc0,
            tau_npg,
            taut_npg,
            core_dens['tauc_g'],
            core_dens['tauct_g'],
        )
"""


def calculate_cider_paw_correction(
    expansion,
    setup,
    D_sp,
    dEdD_sp=None,
    addcoredensity=True,
    a=None,
    separate_ae_ps=False,
    has_cider_contribs=True,
):
    xcc = setup.xc_correction
    rgd = xcc.rgd
    nspins = len(D_sp)

    if has_cider_contribs and expansion.rcalc.mode != "semilocal":
        setup.cider_contribs.set_D_sp(D_sp, setup.xc_correction)

    if addcoredensity:
        nc0_sg = rgd.empty(nspins)
        nct0_sg = rgd.empty(nspins)
        dc0_sg = rgd.empty(nspins)
        dct0_sg = rgd.empty(nspins)
        nc0_sg[:] = sqrt(4 * pi) / nspins * xcc.nc_g
        nct0_sg[:] = sqrt(4 * pi) / nspins * xcc.nct_g
        dc0_sg[:] = sqrt(4 * pi) / nspins * xcc.dc_g
        dct0_sg[:] = sqrt(4 * pi) / nspins * xcc.dct_g
        if xcc.nc_corehole_g is not None and nspins == 2:
            nc0_sg[0] -= 0.5 * sqrt(4 * pi) * xcc.nc_corehole_g
            nc0_sg[1] += 0.5 * sqrt(4 * pi) * xcc.nc_corehole_g
            dc0_sg[0] -= 0.5 * sqrt(4 * pi) * xcc.dc_corehole_g
            dc0_sg[1] += 0.5 * sqrt(4 * pi) * xcc.dc_corehole_g
    else:
        nc0_sg = 0
        nct0_sg = 0
        dc0_sg = 0
        dct0_sg = 0

    D_sLq = np.inner(D_sp, xcc.B_pqL.T)

    res = expansion(rgd, D_sLq, xcc.n_qg, xcc.d_qg, nc0_sg, dc0_sg, ae=True)
    rest = expansion(rgd, D_sLq, xcc.nt_qg, xcc.dt_qg, nct0_sg, dct0_sg, ae=False)

    if expansion.rcalc.mode == "feature":
        if separate_ae_ps:
            return res, rest
        else:
            return res - rest
    elif expansion.rcalc.mode == "debug_nosph":
        return res, rest
    elif expansion.rcalc.mode in ["energy", "energy_v2"]:
        e, dEdD_sqL, x_sbgL = res
        et, dEtdD_sqL, xt_sbgL = rest

        dEdD_sp = np.inner(
            (dEdD_sqL - dEtdD_sqL).reshape((nspins, -1)),
            xcc.B_pqL.reshape((len(xcc.B_pqL), -1)),
        )
        dEdD_sp += setup.cider_contribs.get_dH_sp()

        if addcoredensity:
            ediff = e - et - xcc.e_xc0
        else:
            ediff = e - et
        ndiff = x_sbgL.shape[1] - xt_sbgL.shape[1]
        shape = (x_sbgL.shape[0], ndiff, x_sbgL.shape[2], x_sbgL.shape[3])
        if separate_ae_ps:
            return ediff, dEdD_sp, x_sbgL, np.append(xt_sbgL, np.zeros(shape), axis=1)
        else:
            xdiff = x_sbgL - np.append(xt_sbgL, np.zeros(shape), axis=1)
            return ediff, dEdD_sp, xdiff
    elif expansion.rcalc.mode == "potential":
        dEdD_sqL = res
        dEtdD_sqL = rest
        dEdD_sp = np.inner(
            (dEdD_sqL - dEtdD_sqL).reshape((nspins, -1)),
            xcc.B_pqL.reshape((len(xcc.B_pqL), -1)),
        )
        dEdD_sp += setup.cider_contribs.get_dH_sp()
        return dEdD_sp
    elif expansion.rcalc.mode == "semilocal":
        e, dEdD_sqL = res
        et, dEtdD_sqL = rest

        if dEdD_sp is not None:
            dEdD_sp += np.inner(
                (dEdD_sqL - dEtdD_sqL).reshape((nspins, -1)),
                xcc.B_pqL.reshape((len(xcc.B_pqL), -1)),
            )

        if addcoredensity:
            return e - et - xcc.e_xc0
        else:
            return e - et
    else:
        raise ValueError(
            "Unsupported mode for CIDER PAW correction: {}.".format(
                expansion.rcalc.mode
            )
        )


class CiderRadialExpansion:
    def __init__(self, rcalc, F_sbgr=None, y_sbLg=None, f_in_sph_harm=False):
        self.rcalc = rcalc
        if F_sbgr is not None:
            nspins, nb = F_sbgr.shape[:2]
            if f_in_sph_harm:
                F_sbLg = F_sbgr
                self.F_sbg = np.einsum("sbLg,nL->sbng", F_sbLg, Y_nL).reshape(
                    nspins, nb, -1
                )
            else:
                self.F_sbg = F_sbgr.transpose(0, 1, 3, 2).reshape(nspins, nb, -1)
            self.y_sbg = np.einsum("sbLg,nL->sbng", y_sbLg, Y_nL).reshape(
                nspins, nb, -1
            )
        else:
            self.F_sbg = None
            self.y_sbg = None
        assert self.rcalc.mode in [
            "feature",
            "energy",
            "energy_v2",
            "potential",
            "semilocal",
            "debug_nosph",
        ]

    def __call__(self, rgd, D_sLq, n_qg, dndr_qg, nc0_sg, dnc0dr_sg, ae=True):
        n_sLg = np.dot(D_sLq, n_qg)
        n_sLg[:, 0] += nc0_sg

        dndr_sLg = np.dot(D_sLq, dndr_qg)
        dndr_sLg[:, 0] += dnc0dr_sg

        nspins, Lmax, nq = D_sLq.shape
        dEdD_sqL = np.zeros((nspins, nq, Lmax))

        if self.rcalc.mode == "feature":

            x_sbg = self.rcalc(
                rgd, n_sLg, Y_nL[:, :Lmax], dndr_sLg, rnablaY_nLv[:, :Lmax], ae=ae
            )
            nb = x_sbg.shape[1]
            wx_sbng = x_sbg.reshape(nspins, nb, Y_nL.shape[0], -1)

            return (
                4
                * np.pi
                * np.einsum("sbng,nL->sbgL", wx_sbng, weight_n[:, None] * Y_nL)
            )
        elif self.rcalc.mode == "debug_nosph":
            if self.F_sbg is None:
                feat_xg = self.rcalc(
                    rgd, n_sLg, Y_nL[:, :Lmax], dndr_sLg, rnablaY_nLv[:, :Lmax], ae=ae
                )
            else:
                feat_xg = self.rcalc(
                    rgd,
                    n_sLg,
                    Y_nL[:, :Lmax],
                    dndr_sLg,
                    rnablaY_nLv[:, :Lmax],
                    self.F_sbg,
                    self.y_sbg,
                    ae=ae,
                )
            shape = feat_xg.shape[:-1]
            shape = shape + (Y_nL.shape[0], -1)
            feat_xng = feat_xg.reshape(*shape)
            return feat_xng
        elif self.rcalc.mode == "feat_grid":
            pass
        elif self.rcalc.mode == "semilocal":
            E = 0.0
            for n, Y_L in enumerate(Y_nL[:, :Lmax]):
                w = weight_n[n]
                rnablaY_Lv = rnablaY_nLv[n, :Lmax]
                e_g, dedn_sg, b_vsg, a_sg, dedsigma_xg = self.rcalc(
                    rgd, n_sLg, Y_L, dndr_sLg, rnablaY_Lv
                )
                dEdD_sqL += np.dot(rgd.dv_g * dedn_sg, n_qg.T)[:, :, np.newaxis] * (
                    w * Y_L
                )
                dEdD_sqL += (
                    2
                    * np.dot(rgd.dv_g * dedsigma_xg[::nspins] * a_sg, dndr_qg.T)[
                        :, :, np.newaxis
                    ]
                    * (w * Y_L)
                )
                if nspins == 2:
                    dEdD_sqL += np.dot(
                        rgd.dv_g * dedsigma_xg[1] * a_sg[::-1], dndr_qg.T
                    )[:, :, np.newaxis] * (w * Y_L[:Lmax])
                dedsigma_xg *= rgd.dr_g
                B_vsg = dedsigma_xg[::2] * b_vsg
                if nspins == 2:
                    B_vsg += 0.5 * dedsigma_xg[1] * b_vsg[:, ::-1]
                B_vsq = np.dot(B_vsg, n_qg.T)
                dEdD_sqL += 8 * pi * w * np.inner(rnablaY_Lv, B_vsq.T).T
                E += w * rgd.integrate(e_g)

            return E, dEdD_sqL

        elif self.rcalc.mode == "potential2":

            """
            WARNING TODO: DRAFT CODE
            """
            dedn_sg, b_vsg, a_sg, dedsigma_xg = self.rcalc(
                rgd,
                n_sLg,
                Y_nL[:, :Lmax],
                dndr_sLg,
                rnablaY_nLv[:, :Lmax],
                self.F_sbg,
                self.y_sbg,
                ae=ae,
            )

            nn = Y_nL.shape[0]

            dedn_sng = dedn_sg.reshape(nspins, nn, -1)
            dedsigma_xng = dedsigma_xg.reshape(2 * nspins - 1, nn, -1)
            b_vsng = b_vsg.reshape(3, nspins, nn, -1)
            a_sng = a_sg.reshape(nspins, nn, -1)

            dEdD_snq = np.dot(dedn_sng, n_qg.T)
            dEdD_snq += 2 * np.dot(dedsigma_xng[::nspins] * a_sng, dndr_qg.T)
            if nspins == 2:
                dEdD_snq += np.dot(dedsigma_xng[1] * a_sng[::-1], dndr_qg.T)

            dEdD_sqL = np.einsum(
                "snq,nL->sqL", dEdD_snq, weight_n[:, None] * Y_nL[:, :Lmax]
            )

            # dedsigma_xng *= rgd.dr_g
            dedsigma_xng *= 1.0 / (4 * np.pi * rgd.r_g * rgd.r_g)
            B_vsng = dedsigma_xng[::2] * b_vsng
            if nspins == 2:
                B_vsng += 0.5 * dedsigma_xng[1] * b_vsng[:, ::-1]
            B_vsnq = np.dot(B_vsng, n_qg.T)
            dEdD_sqL += (
                8
                * np.pi
                * np.einsum(
                    "nLv,vsnq->sqL",
                    weight_n[:, None, None] * rnablaY_nLv[:, :Lmax],
                    B_vsnq,
                )
            )

            return dEdD_sqL

        elif self.rcalc.mode == "energy_v2":
            e_g, dedn_sg, dedgrad_svg, v_sbg = self.rcalc(
                rgd,
                n_sLg,
                Y_nL[:, :Lmax],
                dndr_sLg,
                rnablaY_nLv[:, :Lmax],
                self.F_sbg,
                self.y_sbg,
                ae=ae,
            )
            nb = v_sbg.shape[1]
            nn = Y_nL.shape[0]

            dedn_sng = dedn_sg.reshape(nspins, nn, -1)
            dedgrad_svng = dedgrad_svg.reshape(nspins, 3, nn, -1)
            # b_vsng = b_vsg.reshape(3, nspins, nn, -1)
            # a_sng = a_sg.reshape(nspins, nn, -1)

            dedgrad_sng = np.einsum("svng,nv->sng", dedgrad_svng, R_nv)
            # va_sLg = np.einsum("sng,nL->sLg", va_sng, weight_n[:, None] * Y_nL[:, :Lmax])
            dEdD_snq = np.dot(rgd.dv_g * dedn_sng, n_qg.T)
            dEdD_snq += np.dot(rgd.dv_g * dedgrad_sng, dndr_qg.T)
            # dEdD_sqL = np.einsum("sLg,qg->sqL", va_sLg, dndr_qg)
            dEdD_sqL = np.einsum(
                "snq,nL->sqL", dEdD_snq, weight_n[:, None] * Y_nL[:, :Lmax]
            )

            B_svnq = np.dot(dedgrad_svng * rgd.dr_g * rgd.r_g, n_qg.T)
            dEdD_sqL += np.einsum(
                "nLv,svnq->sqL",
                (4 * np.pi) * weight_n[:, None, None] * rnablaY_nLv[:, :Lmax],
                B_svnq,
            )

            E = weight_n.dot(rgd.integrate(e_g.reshape(nn, -1)))
            v_sbgL = np.einsum(
                "sbng,nL->sbgL",
                v_sbg.reshape(nspins, nb, nn, -1),
                weight_n[:, None] * Y_nL,
            )
            return E, dEdD_sqL, 4 * np.pi * v_sbgL

        else:
            mode = self.rcalc.mode
            if mode == "energy":
                e_g, dedn_sg, b_vsg, a_sg, dedsigma_xg, v_sbg = self.rcalc(
                    rgd,
                    n_sLg,
                    Y_nL[:, :Lmax],
                    dndr_sLg,
                    rnablaY_nLv[:, :Lmax],
                    self.F_sbg,
                    self.y_sbg,
                    ae=ae,
                )
                nb = v_sbg.shape[1]
            else:
                dedn_sg, b_vsg, a_sg, dedsigma_xg = self.rcalc(
                    rgd,
                    n_sLg,
                    Y_nL[:, :Lmax],
                    dndr_sLg,
                    rnablaY_nLv[:, :Lmax],
                    self.F_sbg,
                    self.y_sbg,
                    ae=ae,
                )

            nn = Y_nL.shape[0]

            dedn_sng = dedn_sg.reshape(nspins, nn, -1)
            dedsigma_xng = dedsigma_xg.reshape(2 * nspins - 1, nn, -1)
            b_vsng = b_vsg.reshape(3, nspins, nn, -1)
            a_sng = a_sg.reshape(nspins, nn, -1)

            dEdD_snq = np.dot(rgd.dv_g * dedn_sng, n_qg.T)
            dEdD_snq += 2 * np.dot(rgd.dv_g * dedsigma_xng[::nspins] * a_sng, dndr_qg.T)
            if nspins == 2:
                dEdD_snq += np.dot(rgd.dv_g * dedsigma_xng[1] * a_sng[::-1], dndr_qg.T)

            dEdD_sqL = np.einsum(
                "snq,nL->sqL", dEdD_snq, weight_n[:, None] * Y_nL[:, :Lmax]
            )

            dedsigma_xng *= rgd.dr_g
            B_vsng = dedsigma_xng[::2] * b_vsng
            if nspins == 2:
                B_vsng += 0.5 * dedsigma_xng[1] * b_vsng[:, ::-1]
            B_vsnq = np.dot(B_vsng, n_qg.T)
            dEdD_sqL += (
                8
                * np.pi
                * np.einsum(
                    "nLv,vsnq->sqL",
                    weight_n[:, None, None] * rnablaY_nLv[:, :Lmax],
                    B_vsnq,
                )
            )

            if mode == "energy":
                E = weight_n.dot(rgd.integrate(e_g.reshape(nn, -1)))
                v_sbgL = np.einsum(
                    "sbng,nL->sbgL",
                    v_sbg.reshape(nspins, nb, nn, -1),
                    weight_n[:, None] * Y_nL,
                )
                return E, dEdD_sqL, 4 * np.pi * v_sbgL
            else:
                return dEdD_sqL


def vec_radial_gga_vars(rgd, n_sLg, Y_nL, dndr_sLg, rnablaY_nLv):
    nspins = len(n_sLg)

    n_sg = np.dot(Y_nL, n_sLg).transpose(1, 0, 2).reshape(nspins, -1)

    a_sg = np.dot(Y_nL, dndr_sLg).transpose(1, 0, 2)
    b_vsg = np.dot(rnablaY_nLv.transpose(0, 2, 1), n_sLg).transpose(1, 2, 0, 3)
    N = Y_nL.shape[0]
    nx = 2 * nspins - 1

    sigma_xg = rgd.empty((nx, N))
    sigma_xg[::2] = (b_vsg**2).sum(0)
    if nspins == 2:
        sigma_xg[1] = (b_vsg[:, 0] * b_vsg[:, 1]).sum(0)
    sigma_xg[:, :, 1:] /= rgd.r_g[1:] ** 2
    sigma_xg[:, :, 0] = sigma_xg[:, :, 1]
    sigma_xg[::2] += a_sg**2
    if nspins == 2:
        sigma_xg[1] += a_sg[0] * a_sg[1]

    sigma_xg = sigma_xg.reshape(nx, -1)
    a_sg = a_sg.reshape(nspins, -1)
    b_vsg = b_vsg.reshape(3, nspins, -1)

    e_g = rgd.empty(N).reshape(-1)
    dedn_sg = rgd.zeros((nspins, N)).reshape(nspins, -1)
    dedsigma_xg = rgd.zeros((nx, N)).reshape(nx, -1)
    return e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg, a_sg, b_vsg


def vec_radial_gga_vars_v2(rgd, n_sLg, Y_nL, dndr_sLg, rnablaY_nLv):
    nspins = len(n_sLg)

    n_sg = np.dot(Y_nL, n_sLg).transpose(1, 0, 2).reshape(nspins, -1)

    b_vsng = np.dot(rnablaY_nLv.transpose(0, 2, 1), n_sLg).transpose(1, 2, 0, 3)
    b_vsng[..., 1:] /= rgd.r_g[1:]
    b_vsng[..., 0] = b_vsng[..., 1]

    a_sng = np.dot(Y_nL, dndr_sLg).transpose(1, 0, 2)
    b_vsng += R_nv.T[:, None, :, None] * a_sng[None, :, :]
    N = Y_nL.shape[0]

    e_g = rgd.empty(N).reshape(-1)
    e_g[:] = 0
    dedn_sg = rgd.zeros((nspins, N)).reshape(nspins, -1)
    dedgrad_svg = rgd.zeros((nspins, 3, N)).reshape(nspins, 3, -1)
    grad_svg = b_vsng.transpose(1, 0, 2, 3).reshape(nspins, 3, -1)
    return e_g, n_sg, dedn_sg, grad_svg, dedgrad_svg


class CiderRadialFeatureCalculator:
    def __init__(self, xc):
        self.xc = xc
        self.mode = "feature"

    def __call__(self, rgd, n_sLg, Y_nL, dndr_sLg, rnablaY_nLv, ae=True):
        (e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg, a_sg, b_vsg) = vec_radial_gga_vars(
            rgd, n_sLg, Y_nL, dndr_sLg, rnablaY_nLv
        )
        return self.xc.get_paw_atom_contribs(n_sg, sigma_xg, ae=ae)[0]


class DiffSLRadialCalculator:
    def __init__(self, kernel):
        self.kernel = kernel
        self.mode = "semilocal"

    def __call__(self, rgd, n_sLg, Y_L, dndr_sLg, rnablaY_Lv):
        (e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg, a_sg, b_vsg) = radial_gga_vars(
            rgd, n_sLg, Y_L, dndr_sLg, rnablaY_Lv
        )
        self.kernel.calculate(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg)
        return e_g, dedn_sg, b_vsg, a_sg, dedsigma_xg


class CiderRadialEnergyCalculator:
    def __init__(self, xc, a=None, mode="energy"):
        self.xc = xc
        self.a = a
        self.mode = mode
        assert mode in ["energy", "potential"]

    def __call__(self, rgd, n_sLg, Y_nL, dndr_sLg, rnablaY_nLv, F_sbg, y_sbg, ae=True):
        (e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg, a_sg, b_vsg) = vec_radial_gga_vars(
            rgd, n_sLg, Y_nL, dndr_sLg, rnablaY_nLv
        )

        if self.mode == "energy":
            v_sag = self.xc.get_paw_atom_evxc(
                e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg, y_sbg, F_sbg, self.a, ae
            )
            return e_g, dedn_sg, b_vsg, a_sg, dedsigma_xg, v_sag
        else:
            self.xc.get_paw_atom_cider_pot(
                n_sg, dedn_sg, sigma_xg, dedsigma_xg, y_sbg, F_sbg, ae
            )

            return dedn_sg, b_vsg, a_sg, dedsigma_xg


class CiderRadialEnergyCalculator2:
    def __init__(self, xc, a=None):
        self.mode = "energy_v2"
        self.xc = xc
        self.a = a

    def __call__(self, rgd, n_sLg, Y_nL, dndr_sLg, rnablaY_nLv, F_sbg, y_sbg, ae=True):
        (e_g, n_sg, dedn_sg, grad_svg, dedgrad_svg) = vec_radial_gga_vars_v2(
            rgd, n_sLg, Y_nL, dndr_sLg, rnablaY_nLv
        )
        v_sag = self.xc.get_paw_atom_evxc_v2(
            e_g, n_sg, dedn_sg, grad_svg, dedgrad_svg, y_sbg, F_sbg, self.a, ae
        )
        return e_g, dedn_sg, dedgrad_svg, v_sag


class DiffGGA(GGA):
    def __init__(self, kernel, stencil=2):
        super(DiffGGA, self).__init__(kernel, stencil)

    def todict(self):
        return {
            "_cider_type": "DiffGGA",
            "type": self.kernel.type,
            "name": self.kernel.name,
        }

    def initialize(self, density, hamiltonian, wfs):
        super(DiffGGA, self).initialize(density, hamiltonian, wfs)
        for setup in density.setups:
            if setup.xc_correction is None:
                continue
            if not isinstance(setup.xc_correction, DiffPAWXCCorrection):
                setup.xc_correction = DiffPAWXCCorrection.from_setup(setup)

    def calculate_paw_correction(
        self, setup, D_sp, dEdD_sp=None, addcoredensity=True, a=None
    ):
        if setup.xc_correction is None:
            return 0.0
        rcalc = DiffSLRadialCalculator(self.kernel)
        expansion = CiderRadialExpansion(rcalc)
        return calculate_cider_paw_correction(
            expansion, setup, D_sp, dEdD_sp, addcoredensity, a
        )


class DiffMGGA(MGGA):
    def __init__(self, kernel, stencil=2):
        super(DiffMGGA, self).__init__(kernel, stencil)

    def todict(self):
        return {
            "_cider_type": "DiffMGGA",
            "type": self.kernel.type,
            "name": self.kernel.name,
        }

    def initialize(self, density, hamiltonian, wfs):
        super(DiffMGGA, self).initialize(density, hamiltonian, wfs)
        for setup in density.setups:
            if setup.xc_correction is None:
                raise NotImplementedError("SL MGGA not supported with NCPP")
            if not isinstance(setup.xc_correction, DiffPAWXCCorrection):
                setup.xc_correction = DiffPAWXCCorrection.from_setup(
                    setup, build_kinetic=True
                )

    def calculate_paw_correction(
        self, setup, D_sp, dEdD_sp=None, addcoredensity=True, a=None
    ):
        if hasattr(self, "D_sp"):
            raise ValueError
        self.D_sp = D_sp
        self.n = 0
        self.ae = True
        self.xcc = setup.xc_correction
        if self.xcc is None:
            return 0.0
        self.dEdD_sp = dEdD_sp
        self.tau_npg = self.xcc.tau_npg
        self.taut_npg = self.xcc.taut_npg

        rcalc = self.create_mgga_radial_calculator()
        expansion = CiderRadialExpansion(rcalc)
        # The damn thing uses too many 'self' variables to define a clean
        # integrator object.
        E = calculate_cider_paw_correction(
            expansion, setup, D_sp, dEdD_sp, addcoredensity, a
        )
        del self.D_sp, self.n, self.ae, self.xcc, self.dEdD_sp
        return E

    def create_mgga_radial_calculator(self):
        class MockKernel:
            def __init__(self, mgga):
                self.mgga = mgga

            def calculate(self, e_g, n_sg, v_sg, sigma_xg, dedsigma_xg):
                self.mgga.mgga_radial(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg)

        return DiffSLRadialCalculator(MockKernel(self))


class DiffMGGA2(DiffMGGA):
    def initialize(self, density, hamiltonian, wfs):
        super(DiffMGGA, self).initialize(density, hamiltonian, wfs)
        for setup in density.setups:
            if setup.xc_correction is None:
                raise NotImplementedError("SL MGGA not supported with NCPP")
            if not isinstance(setup.xc_correction, DiffPAWXCCorrection):
                setup.xc_correction = DiffPAWXCCorrection.from_setup(
                    setup, build_kinetic=True, ke_order_ng=False
                )



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import yaml
from gpaw.calculator import GPAW
from gpaw.xc.libxc import LibXC

from ciderpress.dft.model_utils import load_cider_model
from ciderpress.gpaw.cider_fft import CiderGGA, CiderMGGA
from ciderpress.gpaw.cider_kernel import CiderGGAHybridKernel, CiderMGGAHybridKernel
from ciderpress.gpaw.cider_paw import CiderGGAPASDW, CiderMGGAPASDW
from ciderpress.gpaw.cider_sl import (
    SLCiderGGA,
    SLCiderGGAHybridWrapper,
    SLCiderMGGA,
    SLCiderMGGAHybridWrapper,
)
from ciderpress.gpaw.interp_paw import DiffGGA, DiffMGGA


def get_cider_functional(
    mlfunc,
    xmix=1.00,
    xkernel="GGA_X_PBE",
    ckernel="GGA_C_PBE",
    mlfunc_format=None,
    use_paw=True,
    pasdw_ovlp_fit=True,
    pasdw_store_funcs=False,
    Nalpha=None,
    qmax=300,
    lambd=1.8,
    _force_nonlocal=False,
):
    """
    Initialize a CIDER surrogate hybrid XC functional of the form
    E_xc = E_x^sl * (1 - xmix) + E_c^sl + xmix * E_x^CIDER
    where E_x^sl is given by the xkernel parameter, E_c^sl is given by the ckernel
    parameter, and E_x^CIDER is the ML exchange energy contains in mlfunc.

    NOTE: Do not use CIDER with ultrasoft pseudopotentials (PAW only).
    At your own risk, you can use CIDER with norm-conserving pseudopotentials,
    but correctness is not guaranteed because the nonlocal features will
    not be correct due to the lack of core electrons.

    NOTE: If the mlfunc is determined to be semilocal, all the
    internal settings are ignored, and a simpler, more efficient
    class is returned to evaluate the semilocal functional.

    Args:
        mlfunc (MappedXC, MappedXC2, str): An ML functional object or a str
            corresponding to the file name of a yaml or joblib file
            containing it.
        xmix (float, 1.00): Mixing parameter for CIDER exchnange.
        xkernel (str, GGA_X_PBE): libxc code str for semi-local X functional
        ckernel (str, GGA_C_PBE): libxc code str for semi-local C functional
        mlfunc_format (str, None): 'joblib' or 'yaml', specifies the format
            of mlfunc if it is a string corresponding to a file name.
            If unspecified, infer from file extension and raise error
            if file type cannot be determined.
        use_paw (bool, True): Whether to compute PAW corrections. This
            should be True unless you plan to use norm-conserving
            pseudopotentials (NCPPs). Note, the use of NCPPs is allowed
            but not advised, as large errors might arise due to the use
            of incorrect nonlocal features.

    The following parameters are only used if use_paw is True. They
    control internal behavior of the PASDW algorithm used to
    evaluate the nonlocal density features in CIDER, but users might
    want to set them to address numerical stability or memory/performance
    tradeoff issues.

    PASDW args:
        pasdw_ovlp_fit (bool, True): Whether to use overlap fitting to
            attempt to improve numerical precision of PASDW projections.
            Default to True. Impact of this parameter should be minor.
        pasdw_store_funcs (bool, False): Whether to store projector
            functions used in the PASDW routine. Defaults to False
            because this can be memory intensive, but if you have the
            space, the computational cost of the atomic corrections
            can be greatly reduced by setting to True.

    The following additional arguments can be set but are considered
    internal parameters and should usually be kept as their defaults,
    which should be reasonable for most cases.

    Internal args:
        Nalpha (int, None):
            Number of interpolation points for the nonlocal feature kernel.
            If None, set automatically based on lambd, qmax, and qmin.
        qmax (float, 300):
            Maximum value of q to use for kernel interpolation on FFT grid.
            Default should be fine for most cases.
        qmin (not currently a parameter, set automatically):
            Mininum value of q to use for kernel interpolation.
            Currently, qmin is set automatically based
            on the minimum regularized value of the kernel exponent.
        lambd (float, 1.8):
            Density of interpolation points. q_alpha=q_0 * lambd**alpha.
            Smaller lambd is more expensive and more precise.

    The following options are for debugging only. Do not set them unless
    you want to test numerical stability issues, as the energies will
    not be variational. Only supported for GGA-type functionals.

    Debug args:
        _force_nonlocal (bool, False): Use nonlocal kernel even if nonlocal
            features are not required. For debugging use.

    Returns:
        A _CiderBase object (specific class depends on the parameters
        provided above, specifically on whether use_paw is True,
        whether the functional is GGA or MGGA form, and whether the
        functional is semi-local or nonlocal.)
    """

    mlfunc = load_cider_model(mlfunc, mlfunc_format)
    if mlfunc.settings.sl_settings.level == "MGGA":
        mlfunc.desc_version = "b"
    else:
        mlfunc.desc_version = "d"

    if mlfunc.desc_version == "b":
        no_nldf = mlfunc.settings.nldf_settings.is_empty
        if no_nldf and not _force_nonlocal:
            # functional is semi-local MGGA
            cider_kernel = SLCiderMGGAHybridWrapper(mlfunc, xmix, xkernel, ckernel)
            return SLCiderMGGA(cider_kernel)
        cider_kernel = CiderMGGAHybridKernel(mlfunc, xmix, xkernel, ckernel)
        if use_paw:
            cls = CiderMGGAPASDW
        else:
            cls = CiderMGGA
    elif mlfunc.desc_version == "d":
        no_nldf = mlfunc.settings.nldf_settings.is_empty
        if no_nldf and not _force_nonlocal:
            # functional is semi-local GGA
            cider_kernel = SLCiderGGAHybridWrapper(mlfunc, xmix, xkernel, ckernel)
            return SLCiderGGA(cider_kernel)
        cider_kernel = CiderGGAHybridKernel(mlfunc, xmix, xkernel, ckernel)
        if use_paw:
            cls = CiderGGAPASDW
        else:
            cls = CiderGGA
    else:
        msg = "Only implemented for b and d version, found {}"
        raise ValueError(msg.format(mlfunc.desc_version))

    if use_paw:
        xc = cls(
            cider_kernel,
            Nalpha=Nalpha,
            lambd=lambd,
            encut=qmax,
            pasdw_ovlp_fit=pasdw_ovlp_fit,
            pasdw_store_funcs=pasdw_store_funcs,
        )
    else:
        xc = cls(cider_kernel, Nalpha=Nalpha, lambd=lambd, encut=qmax)

    return xc


class CiderGPAW(GPAW):
    """
    This class is equivalent to the GPAW calculator object
    provded in the gpaw pacakage, except that it is able to load
    and save CIDER calculations. The GPAW object can run CIDER but
    not save/load CIDER calculations.

    One can also provide CIDER XC functional in dictionary format,
    but this is not advised since one also has to explicitly provide
    by the mlfunc_data parameter, which is the str of the yaml file
    representing the mlfunc object.
    """

    def __init__(
        self,
        restart=None,
        *,
        label=None,
        timer=None,
        communicator=None,
        txt="?",
        parallel=None,
        **kwargs
    ):
        if isinstance(kwargs.get("xc"), dict) and "_cider_type" in kwargs["xc"]:
            if "mlfunc_data" not in kwargs:
                raise ValueError("Must provide mlfunc_data for CIDER.")
            self._mlfunc_data = kwargs.pop("mlfunc_data")
        super(CiderGPAW, self).__init__(
            restart=restart,
            label=label,
            timer=timer,
            communicator=communicator,
            txt=txt,
            parallel=parallel,
            **kwargs,
        )

    def _write(self, writer, mode):
        writer = super(CiderGPAW, self)._write(writer, mode)
        if hasattr(self.hamiltonian.xc, "get_mlfunc_data"):
            mlfunc_data = self.hamiltonian.xc.get_mlfunc_data()
            writer.child("mlfunc").write(mlfunc_data=mlfunc_data)
        return writer

    def initialize(self, atoms=None, reading=False):
        xc = self.parameters.xc
        if isinstance(xc, dict) and "_cider_type" in xc:
            is_cider = "Cider" in xc["_cider_type"]
            if is_cider:
                if reading:
                    self._mlfunc_data = self.reader.mlfunc.get("mlfunc_data")
                xc["mlfunc"] = self._mlfunc_data
            self.parameters.xc = cider_functional_from_dict(xc)
            if is_cider:
                del self._mlfunc_data
        GPAW.initialize(self, atoms, reading)


def cider_functional_from_dict(d):
    """
    This is a function for reading CIDER functionals (and also
    standard functionals using the DiffPAW approach) from a dictionary
    that can be stored in a .gpw file. This is primarily a helper function
    for internal use, as it just reads the functional type and calls
    the from_dict() function of the respective class.

    Args:
        d: XC data

    Returns:
        XC Functional object for use in GPAW.
    """
    cider_type = d.pop("_cider_type")
    if cider_type == "CiderGGAPASDW":
        cls = CiderGGAPASDW
        kcls = CiderGGAHybridKernel
    elif cider_type == "CiderMGGAPASDW":
        cls = CiderMGGAPASDW
        kcls = CiderMGGAHybridKernel
    elif cider_type == "CiderGGA":
        cls = CiderGGA
        kcls = CiderGGAHybridKernel
    elif cider_type == "CiderMGGA":
        cls = CiderMGGA
        kcls = CiderMGGAHybridKernel
    elif cider_type == "SLCiderGGA":
        cls = SLCiderGGA
        kcls = SLCiderGGAHybridWrapper
    elif cider_type == "SLCiderMGGA":
        cls = SLCiderMGGA
        kcls = SLCiderMGGAHybridWrapper
    elif cider_type == "DiffGGA":
        cls = DiffGGA
        kcls = None
    elif cider_type == "DiffMGGA":
        cls = DiffMGGA
        kcls = None
    else:
        raise ValueError("Unsupported functional type")

    if "Cider" in cider_type:
        mlfunc = yaml.load(d["mlfunc"], Loader=yaml.CLoader)
        # kernel_params should be xmix, xkernel, ckernel
        cider_kernel = kcls(mlfunc, **(d["kernel_params"]))
        if "SLCider" in cider_type:
            xc = cls(cider_kernel)
        else:
            # xc_params should have Nalpha, lambd, encut.
            # For PAW, it should also have pasdw_ovlp_fit, pasdw_store_funcs.
            xc = cls(cider_kernel, **(d["xc_params"]))
    else:
        xc = cls(LibXC(d["name"]))

    return xc



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import numpy as np

from ciderpress.dft.lcao_interpolation import LCAOInterpolatorDirect


class LCAONLDFGenerator:
    def __init__(self, plan, ccl, interpolator, grids_indexer):
        """

        Args:
            plan (NLDFAuxiliaryPlan):
            ccl (ConvolutionCollection):
            interpolator (LCAOInterpolator):
            grids_indexer (AtomicGridsIndexer):
        """
        self.plan = plan
        self.ccl = ccl
        self.interpolator = interpolator
        self.grids_indexer = grids_indexer
        self._cache = {s: None for s in range(plan.nspin)}
        nalpha = self.plan.nalpha
        self._uq_buf = np.empty((self.ccl.atco_inp.nao, nalpha))
        self._vq_buf = np.empty((self.ccl.atco_out.nao, self.ccl.num_out))
        self._rlmq_buf = self.grids_indexer.empty_rlmq(nalpha)

    @property
    def natm(self):
        return self.interpolator.atom_coords.shape[0]

    def _perform_fwd_convolution(self, theta_gq, grad_mode=False):
        theta_rlmq = self._rlmq_buf
        theta_uq = self._uq_buf
        conv_vq = self._vq_buf
        conv_vq[:] = 0.0
        theta_uq[:] = 0.0
        theta_rlmq[:] = 0.0
        self.grids_indexer.reduce_angc_ylm_(theta_rlmq, theta_gq, a2y=True)
        self.ccl.atco_inp.convert_rad2orb_(
            theta_rlmq,
            theta_uq,
            self.grids_indexer,
            self.grids_indexer.rad_arr,
            rad2orb=True,
            offset=0,
        )
        self.plan.get_transformed_interpolation_terms(
            theta_uq, i=-1, fwd=True, inplace=True
        )
        self.ccl.multiply_atc_integrals(theta_uq, output=conv_vq, fwd=True)
        if self.plan.nldf_settings.nldf_type != "i":
            self.plan.get_transformed_interpolation_terms(
                conv_vq[:, : self.plan.nalpha], i=0, fwd=False, inplace=True
            )
        if grad_mode:
            return self.interpolator.project_orb2grid(conv_vq), conv_vq.copy()
        else:
            return self.interpolator.project_orb2grid(conv_vq)

    def _perform_bwd_convolution(self, vf_gq):
        vtheta_rlmq = self._rlmq_buf
        vtheta_uq = self._uq_buf
        vconv_vq = self._vq_buf
        vconv_vq[:] = 0.0
        vtheta_uq[:] = 0.0
        self.interpolator.project_grid2orb(vf_gq, f_uq=vconv_vq)
        if self.plan.nldf_settings.nldf_type != "i":
            self.plan.get_transformed_interpolation_terms(
                vconv_vq[:, : self.plan.nalpha], i=0, fwd=True, inplace=True
            )
        self.ccl.multiply_atc_integrals(vconv_vq, output=vtheta_uq, fwd=False)
        self.plan.get_transformed_interpolation_terms(
            vtheta_uq, i=-1, fwd=False, inplace=True
        )
        self.ccl.atco_inp.convert_rad2orb_(
            vtheta_rlmq,
            vtheta_uq,
            self.grids_indexer,
            self.grids_indexer.rad_arr,
            rad2orb=False,
            offset=0,
        )
        vtheta_gq = self.grids_indexer.empty_gq(self.plan.nalpha)
        self.grids_indexer.reduce_angc_ylm_(vtheta_rlmq, vtheta_gq, a2y=False)
        return vtheta_gq

    def get_features(self, rho_in, spin=0, map_grids=True, grad_mode=False):
        # set up arrays
        self._cache[spin] = None
        ngrids_ato = self.grids_indexer.ngrids
        idx_map = self.grids_indexer.idx_map
        ngrids = self.grids_indexer.idx_map.size

        rho_tuple = self.plan.get_rho_tuple(rho_in)
        arg_in_g = self.plan.get_interpolation_arguments(rho_tuple, i=-1)
        func_in_g = self.plan.get_function_to_convolve(rho_tuple)
        arg_in_g, darg_in_g = arg_in_g[0], arg_in_g[1]
        func_in_g, dfunc_in_g = func_in_g[0], func_in_g[1]
        if map_grids:
            arg_g = np.zeros(ngrids_ato)
            func_g = np.zeros(ngrids_ato)
            arg_g[idx_map] = arg_in_g[:ngrids]
            func_g[idx_map] = func_in_g[:ngrids]
        else:
            arg_g = arg_in_g
            func_g = func_in_g
        p_gq, dp_gq = self.plan.get_interpolation_coefficients(arg_g, i=-1)
        theta_gq = p_gq * (func_g * self.grids_indexer.all_weights)[:, None]

        # compute feature convolutions
        f_gq = self._perform_fwd_convolution(theta_gq, grad_mode=grad_mode)
        if grad_mode:
            f_gq, conv_vq = f_gq

        if map_grids:
            rho_out = rho_in
        else:
            rho_out = np.zeros((rho_in.shape[0], ngrids + self.grids_indexer.padding))
            rho_out[:, :ngrids] = rho_in[:, idx_map]
        # do contractions to get features
        feat, dfeat = self.plan.eval_rho_full(
            f_gq,
            rho_out,
            apply_transformation=False,
            spin=spin,
        )
        self._cache[spin] = {
            "rho_in": rho_in,
            "rho_out": rho_out,
            "dfeat": dfeat,
            "p_gq": p_gq,
            "dp_gq": dp_gq,
            "darg_in_g": darg_in_g,
            "func_g": func_g,
            "dfunc_in_g": dfunc_in_g,
        }
        if grad_mode:
            self._cache[spin]["conv_vq"] = conv_vq
        return feat

    def get_features_and_occ_derivs(
        self, rho_in, orb_rho_in_iorb, rho_out=None, orb_rho_out_iorb=None
    ):
        if isinstance(self.interpolator, LCAOInterpolatorDirect):
            raise ValueError("Direct interpolator does not support this function.")
        if rho_out is None:
            assert orb_rho_out_iorb is None
            rho_out = rho_in
            orb_rho_out_iorb = orb_rho_in_iorb
        # set up arrays
        ngrids_ato = self.grids_indexer.ngrids
        idx_map = self.grids_indexer.idx_map
        ngrids = self.grids_indexer.idx_map.size

        rho_tuple = self.plan.get_rho_tuple(rho_in)
        arg_in_g = self.plan.get_interpolation_arguments(rho_tuple, i=-1)
        func_in_g = self.plan.get_function_to_convolve(rho_tuple)
        arg_in_g, darg_in_g = arg_in_g[0], arg_in_g[1]
        func_in_g, dfunc_in_g = func_in_g[0], func_in_g[1]
        arg_g = np.zeros(ngrids_ato)
        func_g = np.zeros(ngrids_ato)
        arg_g[idx_map] = arg_in_g[:ngrids]
        func_g[idx_map] = func_in_g[:ngrids]
        p_gq, dp_gq = self.plan.get_interpolation_coefficients(arg_g, i=-1)
        theta_gq = p_gq * (func_g * self.grids_indexer.all_weights)[:, None]

        # compute feature convolutions
        f_gq = self._perform_fwd_convolution(theta_gq)
        start = 0 if self.plan.nldf_settings.nldf_type == "i" else self.plan.nalpha
        start += len(self.plan.nldf_settings.l0_feat_specs)
        feat = self.plan.eval_rho_full(f_gq, rho_out, apply_transformation=False)[0]
        occd_feats = []

        for iorb in range(orb_rho_in_iorb.shape[0]):
            orb_rho_in = orb_rho_in_iorb[iorb]
            orb_rho_out = orb_rho_out_iorb[iorb]
            occd_sigma_in = 2 * np.einsum("xg,xg", rho_in[1:4], orb_rho_in[1:4])
            occd_arg_in_g = orb_rho_in[0] * darg_in_g[0] + occd_sigma_in * darg_in_g[1]
            occd_func_in_g = (
                orb_rho_in[0] * dfunc_in_g[0] + occd_sigma_in * dfunc_in_g[1]
            )
            if self.plan.nldf_settings.sl_level == "MGGA":
                occd_arg_in_g[:] += orb_rho_in[4] * darg_in_g[2]
                occd_func_in_g[:] += orb_rho_in[4] * dfunc_in_g[2]
            occd_arg_g = np.zeros(ngrids_ato)
            occd_func_g = np.zeros(ngrids_ato)
            occd_arg_g[idx_map] = occd_arg_in_g[:ngrids]
            occd_func_g[idx_map] = occd_func_in_g[:ngrids]
            occd_theta_gq = (
                dp_gq * (func_g * occd_arg_g)[:, None] + p_gq * occd_func_g[:, None]
            )
            occd_theta_gq[:] *= self.grids_indexer.all_weights[:, None]
            occd_f_gq = self._perform_fwd_convolution(occd_theta_gq)
            occd_feat = self.plan.eval_occd_full(
                f_gq,
                rho_out,
                occd_f_gq,
                orb_rho_out,
                apply_transformation=False,
            )
            occd_feats.append(occd_feat)
        return feat, None if len(occd_feats) == 0 else np.stack(occd_feats)

    def get_potential(self, vfeat, spin=0, map_grids=True, grad_mode=False):
        # set up arrays
        if self._cache[spin] is None:
            raise ValueError("Need to call get_features before get_potential")
        cache = self._cache[spin]
        idx_map = self.grids_indexer.idx_map
        ngrids = self.grids_indexer.idx_map.size
        vrho_out = np.zeros_like(cache["rho_out"])
        ngrid_out = vfeat.shape[1]

        # compute feature potential
        vf_gq = np.zeros((vfeat.shape[1], self.interpolator.num_out))
        vf_gq = self.plan.eval_vxc_full(
            vfeat,
            vrho_out,
            cache["dfeat"],
            cache["rho_out"],
            vf=vf_gq,
            spin=spin,
        )
        if map_grids:
            pass
        else:
            vrho_out_tmp = vrho_out
            vrho_out = np.zeros(cache["rho_in"].shape)
            vrho_out[:, idx_map] = vrho_out_tmp[:, :ngrids]
        if grad_mode:
            excsum = self.interpolator.project_orb2grid_grad(cache["conv_vq"], vf_gq)
        vtheta_gq = self._perform_bwd_convolution(vf_gq)
        varg_g = np.einsum("gq,gq->g", cache["dp_gq"], vtheta_gq) * cache["func_g"]
        vfunc_g = np.einsum("gq,gq->g", cache["p_gq"], vtheta_gq)
        varg_g[:] *= self.grids_indexer.all_weights
        if grad_mode:
            cidergg_g = vfunc_g * cache["func_g"]
        vfunc_g[:] *= self.grids_indexer.all_weights

        if map_grids:
            varg_out_g = np.zeros(ngrid_out)
            varg_out_g[:ngrids] = varg_g[idx_map]
            vfunc_out_g = np.zeros(ngrid_out)
            vfunc_out_g[:ngrids] = vfunc_g[idx_map]
        else:
            varg_out_g = varg_g
            vfunc_out_g = vfunc_g
        vrho_out[0] += (
            varg_out_g * cache["darg_in_g"][0] + vfunc_out_g * cache["dfunc_in_g"][0]
        )
        vrho_out[1:4] += (
            (varg_out_g * cache["darg_in_g"][1] + vfunc_out_g * cache["dfunc_in_g"][1])
            * 2
            * cache["rho_in"][1:4]
        )
        if self.plan.nldf_settings.sl_level == "MGGA":
            vrho_out[4] += (
                varg_out_g * cache["darg_in_g"][2]
                + vfunc_out_g * cache["dfunc_in_g"][2]
            )
        # self._cache[spin] = None
        if grad_mode:
            if map_grids:
                raise NotImplementedError("map_grids and grad mode at same time")
            return vrho_out, cidergg_g, excsum
        else:
            return vrho_out



import ctypes

import numpy as np

from ciderpress.lib import load_library

pw_cutil = load_library("libpwutil.so")


def eval_pasdw_funcs(radfuncs_ng, ylm_lg, nlist_i, lmlist_i):
    assert radfuncs_ng.flags.c_contiguous
    assert ylm_lg.flags.c_contiguous
    assert nlist_i.flags.c_contiguous
    assert lmlist_i.flags.c_contiguous
    nlm, ng = ylm_lg.shape
    assert radfuncs_ng.shape[1] == ng
    ni = len(nlist_i)
    assert len(nlist_i) == len(lmlist_i)
    funcs_ig = np.empty((ni, ng), order="C")
    pw_cutil.eval_pasdw_funcs(
        radfuncs_ng.ctypes.data_as(ctypes.c_void_p),
        ylm_lg.ctypes.data_as(ctypes.c_void_p),
        funcs_ig.ctypes.data_as(ctypes.c_void_p),
        nlist_i.ctypes.data_as(ctypes.c_void_p),
        lmlist_i.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(ni),
        ctypes.c_int(ng),
        ctypes.c_int(nlm),
    )
    return funcs_ig


def pasdw_reduce(fn, coefs_i, funcs_ig, augfeat_g, indset):
    assert coefs_i.flags.c_contiguous
    assert funcs_ig.flags.c_contiguous
    assert augfeat_g.flags.c_contiguous
    assert indset.flags.c_contiguous
    ni, ng = funcs_ig.shape
    assert coefs_i.size == ni
    assert indset.shape == (ng, 3)
    assert indset.dtype == np.int32
    n1, n2, n3 = augfeat_g.shape
    fn(
        coefs_i.ctypes.data_as(ctypes.c_void_p),
        funcs_ig.ctypes.data_as(ctypes.c_void_p),
        augfeat_g.ctypes.data_as(ctypes.c_void_p),
        indset.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(ni),
        ctypes.c_int(ng),
        ctypes.c_int(n1),
        ctypes.c_int(n2),
        ctypes.c_int(n3),
    )


def pasdw_reduce_i(coefs_i, funcs_ig, augfeat_g, indset):
    pasdw_reduce(pw_cutil.pasdw_reduce_i, coefs_i, funcs_ig, augfeat_g, indset)


def pasdw_reduce_g(coefs_i, funcs_ig, augfeat_g, indset):
    pasdw_reduce(pw_cutil.pasdw_reduce_g, coefs_i, funcs_ig, augfeat_g, indset)


def _eval_cubic_spline(fn, funcs_ntp, t_g, dt_g):
    nn, nt, _np = funcs_ntp.shape
    assert _np == 4
    ng = t_g.size
    assert dt_g.size == ng
    assert funcs_ntp.flags.c_contiguous
    assert t_g.flags.c_contiguous
    assert dt_g.flags.c_contiguous
    radfuncs_ng = np.empty((nn, ng), order="C")
    fn(
        funcs_ntp.ctypes.data_as(ctypes.c_void_p),
        radfuncs_ng.ctypes.data_as(ctypes.c_void_p),
        t_g.ctypes.data_as(ctypes.c_void_p),
        dt_g.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(nn),
        ctypes.c_int(nt),
        ctypes.c_int(ng),
    )
    return radfuncs_ng


def eval_cubic_spline(funcs_ntp, t_g, dt_g):
    return _eval_cubic_spline(pw_cutil.eval_cubic_spline, funcs_ntp, t_g, dt_g)


def eval_cubic_spline_deriv(funcs_ntp, t_g, dt_g):
    return _eval_cubic_spline(pw_cutil.eval_cubic_spline_deriv, funcs_ntp, t_g, dt_g)


def _eval_cubic_interp(i_g, t_g, c_ip, with_deriv):
    for arr, dtype in zip([i_g, t_g, c_ip], [np.int32, np.float64, np.float64]):
        assert arr.dtype == dtype
        assert arr.flags.c_contiguous
    ng = i_g.size
    ni = c_ip.shape[0]
    assert c_ip.shape[1] == 4
    y_g = np.empty(ng, order="C")
    args = [
        i_g.ctypes.data_as(ctypes.c_void_p),
        t_g.ctypes.data_as(ctypes.c_void_p),
        c_ip.ctypes.data_as(ctypes.c_void_p),
        y_g.ctypes.data_as(ctypes.c_void_p),
    ]
    if with_deriv:
        dy_g = np.empty(ng, order="C")
        args.append(dy_g.ctypes.data_as(ctypes.c_void_p))
        res = (y_g, dy_g)
        fn = pw_cutil.eval_cubic_interp
    else:
        res = y_g
        fn = pw_cutil.eval_cubic_interp_noderiv
    args += [
        ctypes.c_int(ng),
        ctypes.c_int(ni),
    ]
    fn(*args)
    return res


def eval_cubic_interp(i_g, t_g, c_ip):
    return _eval_cubic_interp(i_g, t_g, c_ip, True)


def eval_cubic_interp_noderiv(i_g, t_g, c_ip):
    return _eval_cubic_interp(i_g, t_g, c_ip, False)


def recursive_sph_harm(nlm, rhat_gv):
    n = rhat_gv.shape[0]
    res = np.empty((n, nlm), order="C")
    assert rhat_gv.flags.c_contiguous
    pw_cutil.recursive_sph_harm_vec(
        ctypes.c_int(nlm),
        ctypes.c_int(n),
        rhat_gv.ctypes.data_as(ctypes.c_void_p),
        res.ctypes.data_as(ctypes.c_void_p),
    )
    return res


def recursive_sph_harm_deriv(nlm, rhat_gv):
    n = rhat_gv.shape[0]
    res = np.empty((n, nlm), order="C")
    dres = np.empty((n, 3, nlm), order="C")
    assert rhat_gv.flags.c_contiguous
    pw_cutil.recursive_sph_harm_deriv_vec(
        ctypes.c_int(nlm),
        ctypes.c_int(n),
        rhat_gv.ctypes.data_as(ctypes.c_void_p),
        res.ctypes.data_as(ctypes.c_void_p),
        dres.ctypes.data_as(ctypes.c_void_p),
    )
    return res, dres


def mulexp(F_k, theta_k, k2_k, a, b):
    nk = F_k.size
    assert F_k.size == theta_k.size == k2_k.size
    return pw_cutil.mulexp(
        F_k.ctypes.data_as(ctypes.c_void_p),
        theta_k.ctypes.data_as(ctypes.c_void_p),
        k2_k.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_double(a),
        ctypes.c_double(b),
        ctypes.c_int(nk),
    )



import ctypes

from ciderpress.lib.load import load_library

__all__ = ["load_library", "c_double_p", "c_int_p", "c_null_ptr"]

c_double_p = ctypes.POINTER(ctypes.c_double)
c_int_p = ctypes.POINTER(ctypes.c_int)
c_null_ptr = ctypes.POINTER(ctypes.c_void_p)



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

"""
This class provides tools for normalizing features to have more
desirable behavior under uniform scaling.
"""
from abc import ABC, abstractmethod

import numpy as np

CFC = (3.0 / 10) * (3 * np.pi**2) ** (2.0 / 3)


class FeatNormalizer(ABC):
    def _setup_fwd(self, x, xn):
        if xn is None:
            xn = np.empty_like(x)
        return xn

    def _setup_bwd(self, x, dfdx, dfdrho, dfdinh):
        if dfdx is None:
            dfdx = np.empty_like(x)
        if dfdrho is None:
            dfdrho = np.zeros_like(x)
        if dfdinh is None:
            dfdinh = np.zeros_like(x)
        return dfdx, dfdrho, dfdinh

    @abstractmethod
    def get_usp(self):
        pass

    @abstractmethod
    def fill_fwd(self, x, rho, inh, xn=None):
        """

        Args:
            x:
            rho:
            inh:
            xn: Can be empty

        Returns:

        """

    @abstractmethod
    def fill_bwd(self, dfdxn, x, rho, inh, dfdx=None, dfdrho=None, dfdinh=None):
        """

        Args:
            dfdxn:
            x:
            rho:
            inh:
            dfdx: Can be empty
            dfdrho: Must be initialized because it is added to
            dfdinh: Must be initialized because it is added to

        Returns:

        """

    @abstractmethod
    def get_normed_feature_deriv(self, x, rho, inh, dx, drho, dinh):
        pass

    @abstractmethod
    def get_ueg(self, rho=1.0):
        pass


class ConstantNormalizer(FeatNormalizer):
    def __init__(self, const):
        self.const = const

    def get_usp(self):
        return 0

    def fill_fwd(self, x, rho, inh, xn=None):
        xn = self._setup_fwd(x, xn)
        xn[:] = x * self.const
        return xn

    def fill_bwd(self, dfdxn, x, rho, inh, dfdx=None, dfdrho=None, dfdinh=None):
        dfdx, dfdrho, dfdinh = self._setup_bwd(x, dfdx, dfdrho, dfdinh)
        dfdx[:] = dfdxn * self.const
        return dfdx, dfdrho, dfdinh

    def get_normed_feature_deriv(self, x, rho, inh, dx, drho, dinh):
        return dx * self.const

    def get_ueg(self, rho=1.0):
        return self.const


class DensityNormalizer(FeatNormalizer):
    def __init__(self, const, power):
        self.const = const
        self.power = power

    def get_usp(self):
        return 3 * self.power

    def fill_fwd(self, x, rho, inh, xn=None):
        xn = self._setup_fwd(x, xn)
        xn[:] = x * self.const * rho**self.power
        return xn

    def fill_bwd(self, dfdxn, x, rho, inh, dfdx=None, dfdrho=None, dfdinh=None):
        dfdx, dfdrho, dfdinh = self._setup_bwd(x, dfdx, dfdrho, dfdinh)
        fac = self.const * rho**self.power
        dfdx[:] = dfdxn * fac
        dfdrho[:] += dfdxn * self.power * fac * x / rho
        return dfdx, dfdrho, dfdinh

    def get_normed_feature_deriv(self, x, rho, inh, dx, drho, dinh):
        fac = self.const * rho**self.power
        return fac * (dx + self.power * x * drho / rho)

    def get_ueg(self, rho=1.0):
        return self.const * rho**self.power


class InhomogeneityNormalizer(FeatNormalizer):
    def __init__(self, const1, const2, power):
        self.const1 = const1
        self.const2 = const2
        self.power = power

    def get_usp(self):
        return 0.0

    def fill_fwd(self, x, rho, inh, xn=None):
        xn = self._setup_fwd(x, xn)
        inh = 1 + self.const2 * inh
        xn[:] = x * self.const1 * inh**self.power
        return xn

    def fill_bwd(self, dfdxn, x, rho, inh, dfdx=None, dfdrho=None, dfdinh=None):
        dfdx, dfdrho, dfdinh = self._setup_bwd(x, dfdx, dfdrho, dfdinh)
        inh = 1 + self.const2 * inh
        fac = self.const1 * inh**self.power
        dfdx[:] = dfdxn * fac
        dfdinh[:] += (
            dfdxn * x * self.const1 * self.const2 * self.power * inh ** (self.power - 1)
        )
        return dfdx, dfdrho, dfdinh

    def get_normed_feature_deriv(self, x, rho, inh, dx, drho, dinh):
        inh = 1 + self.const2 * inh
        fac = self.const1 * inh**self.power
        return fac * dx + x * self.const1 * self.const2 * dinh * self.power * inh ** (
            self.power - 1
        )

    def get_ueg(self, rho=1.0):
        return self.const1


class GeneralNormalizer(FeatNormalizer):
    def __init__(self, const1, const2, power1, power2):
        self.const1 = const1
        self.const2 = const2
        self.power1 = power1
        self.power2 = power2

    def get_usp(self):
        return 3 * self.power1

    def fill_fwd(self, x, rho, inh, xn=None):
        xn = self._setup_fwd(x, xn)
        fac = self.const1 * rho**self.power1
        inh = 1 + self.const2 * inh
        xn[:] = x * fac * inh**self.power2
        return xn

    def fill_bwd(self, dfdxn, x, rho, inh, dfdx=None, dfdrho=None, dfdinh=None):
        dfdx, dfdrho, dfdinh = self._setup_bwd(x, dfdx, dfdrho, dfdinh)
        fac1 = self.const1 * rho**self.power1
        inh = 1 + self.const2 * inh
        fac2 = inh**self.power2
        dfdx[:] = dfdxn * fac1 * fac2
        dfdrho[:] += dfdxn * x * fac2 * fac1 * self.power1 / rho
        dfdinh[:] += (
            dfdxn * x * fac1 * self.power2 * self.const2 * inh ** (self.power2 - 1)
        )
        return dfdx, dfdrho, dfdinh

    def get_normed_feature_deriv(self, x, rho, inh, dx, drho, dinh):
        fac1 = self.const1 * rho**self.power1
        inh = 1 + self.const2 * inh
        fac2 = inh**self.power2
        res = dx * fac1 * fac2
        res += x * fac2 * fac1 * self.power1 / rho * drho
        res += x * fac1 * self.power2 * self.const2 * inh ** (self.power2 - 1) * dinh
        return res

    def get_ueg(self, rho=1.0):
        return self.const1 * rho**self.power1


def get_invariant_normalizer_from_exponent_params(power, a0, tau_mul):
    tau_fac = tau_mul * 1.2 * (6 * np.pi**2) ** (2.0 / 3) / np.pi
    B = np.pi / 2 ** (2.0 / 3) * (a0 - tau_fac)
    C = np.pi / 2 ** (2.0 / 3) * tau_fac  # / CFC
    const2 = C / B
    const1 = B**power
    return InhomogeneityNormalizer(const1, const2, power)


def get_normalizer_from_exponent_params(rho_pow, exp_pow, a0, tau_mul, gga=False):
    power1 = rho_pow + 2.0 / 3 * exp_pow
    power2 = exp_pow
    tau_fac = tau_mul * 1.2 * (6 * np.pi**2) ** (2.0 / 3) / np.pi
    if gga:
        B = np.pi / 2 ** (2.0 / 3) * a0
    else:
        B = np.pi / 2 ** (2.0 / 3) * (a0 - tau_fac)
    C = np.pi / 2 ** (2.0 / 3) * tau_fac  # / CFC
    const2 = C / B
    const1 = B**exp_pow
    return GeneralNormalizer(const1, const2, power1, power2)


class FeatNormalizerList:
    def __init__(self, normalizers, slmode="npa", cutoff=1e-10):  # , i0=3, i1=None):
        """
        Args:
            normalizers (list[FeatNormalizer or None]):
            list of feature normalizers
        """
        self.slmode = slmode
        self._normalizers = normalizers
        self.cutoff = cutoff
        # self.i0 = i0
        # self.i1 = self.nfeat + i0 if i1 is None else i1
        # if self.i1 - self.i0 != self.nfeat:
        #    raise ValueError("Index range must correspond to number"
        #                     "of normalizers provided.")

    @property
    def nfeat(self):
        return len(self._normalizers)

    def get_usps(self):
        usps = []
        for n in self._normalizers:
            if n is None:
                usps.append(0)
            else:
                usps.append(n.get_usp())
        return usps

    def __getitem__(self, item):
        return self._normalizers[item]

    def __setitem__(self, key, value):
        raise RuntimeError("Cannot set element of normalizer list")

    def _check_shape(self, x, ndim=3):
        assert ndim == 3 or ndim == 2, "Internal error"
        if x.ndim != ndim:
            raise ValueError("Expected {}-dimensional array".format(ndim))
        if x.shape[-2] != self.nfeat:
            raise ValueError("Array must have size nfeat")

    def _get_rho_and_inh(self, X0T):
        rho_term = X0T[:, 0]
        grad_term = X0T[:, 1]
        if self.slmode == "npa":
            tau_term = X0T[:, 2]
            inh_term = tau_term + 5.0 / 3 * grad_term
        elif self.slmode == "nst":
            tau_term = X0T[:, 2]
            inh_term = tau_term / (CFC * rho_term ** (5.0 / 3))
        elif self.slmode == "np":
            inh_term = 5.0 / 3 * grad_term
        else:
            inh_term = grad_term / (8 * CFC * rho_term ** (8.0 / 3))
        rho_term = np.maximum(rho_term, self.cutoff)
        return rho_term, inh_term

    def _get_drho_and_dinh(self, X0T, DX0T):
        rho = X0T[0]
        drho = DX0T[0]
        grad = X0T[1]
        dgrad = DX0T[1]
        if self.slmode == "npa":
            tau = X0T[2]
            dtau = DX0T[2]
            dinh = dtau + 5.0 / 3 * dgrad
        elif self.slmode == "nst":
            tau = X0T[2]
            dtau = DX0T[2]
            dinh = dtau / (CFC * rho ** (5.0 / 3))
            dinh -= 5.0 / 3 * tau / (CFC * rho ** (8.0 / 3)) * drho
        elif self.slmode == "np":
            dinh = 5.0 / 3 * dgrad
        else:
            dinh = dgrad / (8 * CFC * rho ** (8.0 / 3))
            dinh -= 8.0 / 3 * grad / (8 * CFC * rho ** (11.0 / 3)) * drho
        return drho, dinh

    def get_normalized_feature_vector(self, X0T):
        # This function should be in the settings class, rather than
        # a plan, because it is called from models, which do not
        # require knowledge of the plan.
        self._check_shape(X0T)
        X0TN = np.empty_like(X0T)
        rho_term, inh_term = self._get_rho_and_inh(X0T)
        nfeat = self.nfeat
        for i in range(nfeat):
            if self[i] is not None:
                self[i].fill_fwd(X0T[:, i], rho_term, inh_term, xn=X0TN[:, i])
            else:
                X0TN[:, i] = X0T[:, i]
        return X0TN

    def get_derivative_of_normed_features(self, X0T, DX0T):
        self._check_shape(X0T, ndim=2)
        self._check_shape(DX0T, ndim=2)
        DX0TN = np.empty_like(DX0T)
        rho, inh = self._get_rho_and_inh(X0T[None, :])
        rho, inh = rho[0], inh[0]
        drho, dinh = self._get_drho_and_dinh(X0T, DX0T)
        for i in range(self.nfeat):
            if self[i] is not None:
                DX0TN[i] = self[i].get_normed_feature_deriv(
                    X0T[i], rho, inh, DX0T[i], drho, dinh
                )
            else:
                DX0TN[i] = DX0T[i]
        return DX0TN

    def get_derivative_wrt_unnormed_features(self, X0T, df_dX0TN):
        self._check_shape(X0T)
        self._check_shape(df_dX0TN)
        df_dX0T = np.empty_like(X0T)
        rho_term, inh_term = self._get_rho_and_inh(X0T)
        dfdrho = np.zeros_like(rho_term)
        dfdinh = np.zeros_like(inh_term)
        cond = rho_term < self.cutoff
        rho_term[cond] = self.cutoff
        nfeat = self.nfeat
        for i in range(nfeat):
            if self[i] is not None:
                self[i].fill_bwd(
                    df_dX0TN[:, i],
                    X0T[:, i],
                    rho_term,
                    inh_term,
                    dfdx=df_dX0T[:, i],
                    dfdrho=dfdrho,
                    dfdinh=dfdinh,
                )
            else:
                df_dX0T[:, i] = df_dX0TN[:, i]
        df_dX0T[:, 0] += dfdrho
        if self.slmode == "npa":
            df_dX0T[:, 1] += 5.0 / 3 * dfdinh
            df_dX0T[:, 2] += dfdinh
        elif self.slmode == "nst":
            df_dX0T[:, 0] -= dfdinh * 5.0 / 3 * inh_term / rho_term
            df_dX0T[:, 2] += dfdinh / (CFC * rho_term ** (5.0 / 3))
        elif self.slmode == "np":
            df_dX0T[:, 1] += 5.0 / 3 * dfdinh
        else:
            df_dX0T[:, 0] -= dfdinh * 8.0 / 3 * inh_term / rho_term
            df_dX0T[:, 1] += dfdinh / (8 * CFC * rho_term ** (8.0 / 3))
        for s in range(cond.shape[0]):
            df_dX0T[s, ..., cond[s]] = 0.0
        return df_dX0T

    def ueg_vector(self, rho=1.0):
        norms = []
        for n in self._normalizers:
            if n is None:
                norms.append(1.0)
            else:
                norms.append(n.get_ueg(rho))
        return np.array(norms)
